# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import pytest

from ..annotation.antibody import Antibody
from ..annotation.regions import (
    IMGT_REGION_END_POSITIONS_AA,
    IMGT_REGION_END_POSITIONS_NT,
    IMGT_REGION_START_POSITIONS_AA,
    IMGT_REGION_START_POSITIONS_NT,
    get_region_sequence,
)

# =============================================
#                  FIXTURES
# =============================================


@pytest.fixture
def antibody():
    """Create a basic Antibody object with logging methods."""
    return Antibody(sequence_id="test_sequence")


@pytest.fixture
def aligned_sequence():
    """
    A sample aligned sequence spanning all antibody regions.
    This represents a V region with a few gaps aligned to the germline.
    """
    # FWR1 (positions 1-78)
    fwr1 = "CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"
    # CDR1 (positions 79-114), with a single A>T mutation immediately after the IMGT gap
    cdr1 = "CACTTCACCGGCTTGCACTGGGAGTCAGTA"
    # FWR2 (positions 115-165)
    fwr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAGCGGATCAACCCTAACAGT"
    # CDR2 (positions 166-195), with a single codon (AAA) insertion in the IMGT gap
    cdr2 = "GGGGGCAAAACAAACTATGCACAGAAG"
    # FWR3 (positions 196-312), with a single codon deletion immediately after the IMGT gap
    fwr3 = "TTCCAGGGCAGAGTCACCATGACCACAGACACATCCGAT---AGCACAGCCTACATGGAGCTGAGCAGCCTGAGATCTGAGGACACGGCCGTGTATTACTGTGCG"
    # CDR3 (positions 313-351)
    cdr3 = "AGAGTGGGCAACTGGGGCCAGGGTACCTTTGACTACTGG"
    # FWR4 (positions 352-387)
    fwr4 = "GGCCAGGGAACCCTGGTCACCGTCTCCTCAGGTAAG"

    return fwr1 + cdr1 + fwr2 + cdr2 + fwr3 + cdr3 + fwr4


@pytest.fixture
def aligned_germline():
    """
    A sample aligned germline sequence spanning all antibody regions.
    Matches aligned_sequence but with gaps where the sequence has insertions,
    and nucleotides where the sequence has deletions.
    """
    # FWR1 (positions 1-78)
    fwr1 = "CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"
    # CDR1 (positions 79-114)
    cdr1 = "CACTTCACCGGCATGCACTGGGAGTCAGTA"
    # FWR2 (positions 115-165)
    fwr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAGCGGATCAACCCTAACAGT"
    # CDR2 (positions 166-195)
    cdr2 = "GGGGGC---ACAAACTATGCACAGAAG"
    # FWR3 (positions 196-312)
    fwr3 = "TTCCAGGGCAGAGTCACCATGACCACAGACACATCCGATACGAGCACAGCCTACATGGAGCTGAGCAGCCTGAGATCTGAGGACACGGCCGTGTATTACTGTGCG"
    # CDR3 (positions 313-351)
    cdr3 = "AGAGTGGGCAACTGGGGCCAGGGTACCTTTGACTACTGG"
    # FWR4 (positions 352-387)
    fwr4 = "GGCCAGGGAACCCTGGTCACCGTCTCCTCAGGTAAG"

    return fwr1 + cdr1 + fwr2 + cdr2 + fwr3 + cdr3 + fwr4


@pytest.fixture
def gapped_germline():
    """
    A sample IMGT-gapped germline sequence spanning all antibody regions.
    This represents the germline with IMGT-defined gaps (periods) in codon-length groups.
    """
    # IMGT gaps are typically inserted at specific regions in multiples of three to maintain the
    # reading frame. The key IMGT insertion points include:
    # - FWR1: Insertion after codon 6 (position 18)
    # - FWR2: Insertion after codon 45 (position 135) and sometimes another after codon 47
    # - CDR2: Variable insertions to align conserved anchor residues
    # - FWR3: Insertions to maintain alignment of conserved residues

    # FWR1 (positions 1-78) with gaps after position 18 (codon 6)
    fwr1 = "CAGGTTCAGCTG...GTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"  # Codons 1-4
    # CDR1 (positions 79-114)
    cdr1 = "CACTTCACCGGC......ATGCACTGGGAGTCAGTA"
    # FWR2 (positions 115-165) with gaps after position 135 (codon 45)
    fwr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAG...CGGATCAACCCTAACAGT"
    # CDR2 (positions 166-195) with gaps to align conserved residues
    cdr2 = "GGGGGC......ACAAACTATGCACAGAAG"
    # FWR3 (positions 196-312) with a gap to maintain alignment
    fwr3 = "TTCCAGGGCAGAGTCACCATGACCACAGACACATCCGAT............ACGAGCACAGCCTACATGGAGCTGAGCAGCCTGAGATCTGAGGACACGGCCGTGTATTACTGTGCG"
    # CDR3 (positions 313-351) often highly variable but we'll keep it simple
    cdr3 = "AGAGTGGGCAACTGGGGCCAGGGTACCTTTGACTACTGG"
    # FWR4 (positions 352-387)
    fwr4 = "GGCCAGGGAACCCTGGTCACCGTCTCCTCAGGTAAG"

    return fwr1 + cdr1 + fwr2 + cdr2 + fwr3 + cdr3 + fwr4


@pytest.fixture
def truncated_aligned_sequence():
    """
    A truncated aligned sequence (missing part of the 5' end).
    Starts at position 40 (in FWR1).
    """
    # Partial FWR1 (starting at position 40)
    partial_fwr1 = (
        "---------------------------------------CCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"
    )
    # CDR1 (positions 79-114)
    cdr1 = "CACTTCACCGGCATGCACTGGGAGTCAGTA"
    # FWR2 (positions 115-165)
    fwr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAGCGGATCAACCCTAACAGT"
    # CDR2 (positions 166-195)
    cdr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAG---CGGATCAACCCTAACAGT"
    # FWR3 (positions 196-312)
    fwr3 = "TTCCAGGGCAGAGTCACCATGACCACAGACACATCCGATACGAGCACAGCCTACATGGAGCTGAGCAGCCTGAGATCTGAGGACACGGCCGTGTATTACTGTGCG"
    # CDR3 (positions 313-351)
    cdr3 = "AGAGTGGGCAACTGGGGCCAGGGTACCTTTGACTACTGG"
    # FWR4 (positions 352-387)
    fwr4 = "GGCCAGGGAACCCTGGTCACCGTCTCCTCAGGTAAG"

    return partial_fwr1 + cdr1 + fwr2 + cdr2 + fwr3 + cdr3 + fwr4


@pytest.fixture
def truncated_aligned_germline():
    """
    The corresponding aligned germline for the truncated sequence.
    This preserves the full germline, with the sequence being truncated.
    """
    # Partial FWR1 (starting at position 40)
    fwr1 = "CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"
    # CDR1 (positions 79-114)
    cdr1 = "CACTTCACCGGCATGCACTGGGAGTCAGTA"
    # FWR2 (positions 115-165)
    fwr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAGCGGATCAACCCTAACAGT"
    # CDR2 (positions 166-195)
    cdr2 = "GTGCGACAGGCCCCTGGACAAGGGCTTGAG---CGGATCAACCCTAACAGT"
    # FWR3 (positions 196-312)
    fwr3 = "TTCCAGGGCAGAGTCACCATGACCACAGACACATCCGATACGAGCACAGCCTACATGGAGCTGAGCAGCCTGAGATCTGAGGACACGGCCGTGTATTACTGTGCG"
    # CDR3 (positions 313-351)
    cdr3 = "AGAGTGGGCAACTGGGGCCAGGGTACCTTTGACTACTGG"
    # FWR4 (positions 352-387)
    fwr4 = "GGCCAGGGAACCCTGGTCACCGTCTCCTCAGGTAAG"

    return fwr1 + cdr1 + fwr2 + cdr2 + fwr3 + cdr3 + fwr4


@pytest.fixture
def mock_region_data():
    """
    Mock region start/end positions for testing.
    """
    return {
        "fwr1": {"start_nt": 1, "end_nt": 78, "start_aa": 1, "end_aa": 26},
        "cdr1": {"start_nt": 79, "end_nt": 114, "start_aa": 27, "end_aa": 38},
        "fwr2": {"start_nt": 115, "end_nt": 165, "start_aa": 39, "end_aa": 55},
        "cdr2": {"start_nt": 166, "end_nt": 195, "start_aa": 56, "end_aa": 65},
        "fwr3": {"start_nt": 196, "end_nt": 312, "start_aa": 66, "end_aa": 104},
        "cdr3": {"start_nt": 313, "end_nt": 351, "start_aa": 105, "end_aa": 117},
        "fwr4": {"start_nt": 352, "end_nt": 387, "start_aa": 118, "end_aa": 129},
    }


# =============================================
#           REGION SEQUENCE TESTS
# =============================================


def test_get_region_sequence_fwr1(
    aligned_sequence, aligned_germline, gapped_germline, antibody
):
    """Test getting FWR1 region sequence."""
    region = "fwr1"
    result = get_region_sequence(
        region=region,
        aligned_sequence=aligned_sequence,
        aligned_germline=aligned_germline,
        gapped_germline=gapped_germline,
        germline_start=1,
        ab=antibody,
    )
    expected_fwr1 = (
        "CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"
    )
    assert result == expected_fwr1
    # Verify the sequence is not empty
    assert len(result) > 0
    # Verify no gaps in the result
    assert "-" not in result


def test_get_region_sequence_aa(
    aligned_sequence, aligned_germline, gapped_germline, antibody
):
    """Test getting a region sequence as amino acids."""
    region = "fwr1"
    result = get_region_sequence(
        region=region,
        aligned_sequence=aligned_sequence,
        aligned_germline=aligned_germline,
        gapped_germline=gapped_germline,
        germline_start=1,
        ab=antibody,
        aa=True,
    )
    expected_fwr1 = "CAGGTTCAGCTGGTGCAGTCTGG"
    assert result == expected_fwr1
    # Verify the sequence is not empty
    assert len(result) > 0
    # Verify no gaps in the result
    assert "-" not in result


def test_get_region_sequence_truncated(
    truncated_aligned_sequence, truncated_aligned_germline, gapped_germline, antibody
):
    """Test getting a region sequence from a truncated sequence."""
    region = "fwr1"
    result = get_region_sequence(
        region=region,
        aligned_sequence=truncated_aligned_sequence,
        aligned_germline=truncated_aligned_germline,
        gapped_germline=gapped_germline,
        germline_start=1,
        ab=antibody,
    )
    # The sequence is truncated at the 5' end, so we expect a partial FWR1
    expected_partial_fwr1 = "CCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT"
    assert result == expected_partial_fwr1
    assert len(result) > 0
    assert "-" not in result


def test_get_region_sequence_missing_region(
    aligned_sequence, aligned_germline, gapped_germline, antibody
):
    """Test getting a region that's missing from the sequence."""
    # Set germline_start high enough that the region is entirely missing
    germline_start = 400  # Well beyond any regions in our test data
    region = "fwr1"
    result = get_region_sequence(
        region=region,
        aligned_sequence=aligned_sequence,
        aligned_germline=aligned_germline,
        gapped_germline=gapped_germline,
        germline_start=germline_start,
        ab=antibody,
    )
    # The result should be an empty string for a missing region
    assert result == ""


def test_all_regions(
    aligned_sequence, aligned_germline, gapped_germline, antibody, mock_region_data
):
    """Test getting all regions from the sequence."""
    regions = ["fwr1", "cdr1", "fwr2", "cdr2", "fwr3"]

    # Expected sequences for each region
    expected_sequences = {
        "fwr1": "CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCT",
        "cdr1": "CACTTCACCGGCTTGCACTGGGAGTCAGTA",
        "fwr2": "GTGCGACAGGCCCCTGGACAAGGGCTTGAGCGGATCAACCCTAACAGT",
        "cdr2": "GGGGGCAAAACAAACTATGCACAGAAG",
        "fwr3": "TTCCAGGGCAGAGTCACCATGACCACAGACACATCCGATAGCACAGCCTACATGGAGCTGAGCAGCCTGAGATCTGAGGACACGGCCGTGTATTACTGTGCG",
        # "cdr3": "AGAGTGGGCAACTGGGGCCAGGGTACCTTTGACTACTGG",
        # "fwr4": "GGCCAGGGAACCCTGGTCACCGTCTCCTCAGGTAAG",
    }

    for region in regions:
        result = get_region_sequence(
            region=region,
            aligned_sequence=aligned_sequence,
            aligned_germline=aligned_germline,
            gapped_germline=gapped_germline,
            germline_start=1,
            ab=antibody,
        )
        # Compare with expected sequence
        assert result == expected_sequences[region], f"Failed for region {region}"
        # Basic validation - result should be a string and shouldn't contain gaps
        assert isinstance(result, str)
        assert "-" not in result


# =============================================
#         IMGT REGION CONSTANTS TESTS
# =============================================


def test_imgt_region_constants_consistency(mock_region_data):
    """Test that IMGT region constants are consistent with each other."""
    regions = ["fwr1", "cdr1", "fwr2", "cdr2", "fwr3", "cdr3", "fwr4"]

    for region in regions:
        # Check that each region has start and end positions defined
        assert region in IMGT_REGION_START_POSITIONS_NT
        assert region in IMGT_REGION_END_POSITIONS_NT
        assert region in IMGT_REGION_START_POSITIONS_AA
        assert region in IMGT_REGION_END_POSITIONS_AA

        # Verify that start positions are always less than end positions
        assert (
            IMGT_REGION_START_POSITIONS_NT[region]
            <= IMGT_REGION_END_POSITIONS_NT[region]
        )
        assert (
            IMGT_REGION_START_POSITIONS_AA[region]
            <= IMGT_REGION_END_POSITIONS_AA[region]
        )

        # Compare with mock data
        assert (
            IMGT_REGION_START_POSITIONS_NT[region]
            == mock_region_data[region]["start_nt"]
        )
        assert (
            IMGT_REGION_END_POSITIONS_NT[region] == mock_region_data[region]["end_nt"]
        )
        assert (
            IMGT_REGION_START_POSITIONS_AA[region]
            == mock_region_data[region]["start_aa"]
        )
        assert (
            IMGT_REGION_END_POSITIONS_AA[region] == mock_region_data[region]["end_aa"]
        )


def test_region_boundaries_consistency():
    """Test that region boundaries are consistent (end of one region + 1 = start of next region)."""
    # For nucleotides
    assert (
        IMGT_REGION_END_POSITIONS_NT["fwr1"] + 1
        == IMGT_REGION_START_POSITIONS_NT["cdr1"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_NT["cdr1"] + 1
        == IMGT_REGION_START_POSITIONS_NT["fwr2"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_NT["fwr2"] + 1
        == IMGT_REGION_START_POSITIONS_NT["cdr2"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_NT["cdr2"] + 1
        == IMGT_REGION_START_POSITIONS_NT["fwr3"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_NT["fwr3"] + 1
        == IMGT_REGION_START_POSITIONS_NT["cdr3"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_NT["cdr3"] + 1
        == IMGT_REGION_START_POSITIONS_NT["fwr4"]
    )

    # For amino acids
    assert (
        IMGT_REGION_END_POSITIONS_AA["fwr1"] + 1
        == IMGT_REGION_START_POSITIONS_AA["cdr1"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_AA["cdr1"] + 1
        == IMGT_REGION_START_POSITIONS_AA["fwr2"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_AA["fwr2"] + 1
        == IMGT_REGION_START_POSITIONS_AA["cdr2"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_AA["cdr2"] + 1
        == IMGT_REGION_START_POSITIONS_AA["fwr3"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_AA["fwr3"] + 1
        == IMGT_REGION_START_POSITIONS_AA["cdr3"]
    )
    assert (
        IMGT_REGION_END_POSITIONS_AA["cdr3"] + 1
        == IMGT_REGION_START_POSITIONS_AA["fwr4"]
    )


def test_aa_nt_positions_consistent():
    """Test that amino acid positions are consistent with nucleotide positions (nt pos = aa pos * 3)."""
    regions = ["fwr1", "cdr1", "fwr2", "cdr2", "fwr3", "cdr3", "fwr4"]

    for region in regions:
        # Check if AA start position * 3 - 2 = NT start position
        # This accounts for the 3 nucleotides per amino acid, where the first nucleotide
        # of the codon corresponds to the amino acid position
        assert (
            IMGT_REGION_START_POSITIONS_NT[region]
            == (IMGT_REGION_START_POSITIONS_AA[region] * 3) - 2
        )

        # For end positions, the relationship is not exact due to possible partial codons
        # But the relationship should be approximately true
        aa_end_nt_pos = IMGT_REGION_END_POSITIONS_AA[region] * 3
        # End position should be within 2 nucleotides of the expected position
        assert abs(IMGT_REGION_END_POSITIONS_NT[region] - aa_end_nt_pos) <= 2


def test_edge_cases(antibody):
    """Test edge cases for the get_region_sequence function."""
    # Empty sequences
    result = get_region_sequence(
        region="fwr1",
        aligned_sequence="",
        aligned_germline="",
        gapped_germline="",
        germline_start=0,
        ab=antibody,
    )
    assert result == ""

    # Sequences with only gaps
    result = get_region_sequence(
        region="fwr1",
        aligned_sequence="-----",
        aligned_germline="-----",
        gapped_germline=".....",
        germline_start=0,
        ab=antibody,
    )
    assert result == ""



import os
import sys

data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "reference")

# if sys.version_info[0] > 2:
#     import pickle
#     bnab_hc_antibodies_file = 'bnab_hc_antibodies.pkl'
#     bnab_lc_antibodies_file = 'bnab_lc_antibodies.pkl'
# else:
#     import cPickle as pickle
#     bnab_hc_antibodies_file = 'bnab_hc_antibodies.pkl'
#     bnab_lc_antibodies_file = 'bnab_lc_antibodies.pkl'

# with open(os.path.join(data_dir, bnab_hc_antibodies_file), 'rb') as f:
#     REFERENCE_BNAB_HC_ANTIBODIES = pickle.load(f)

# with open(os.path.join(data_dir, bnab_lc_antibodies_file), 'rb') as f:
#     REFERENCE_BNAB_LC_ANTIBODIES = pickle.load(f)



import warnings

from Bio import BiopythonWarning

warnings.simplefilter("ignore", BiopythonWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

from . import gl, pp, tl
from .core.abstar import run
from .version import __version__



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import multiprocessing as mp
import os
import shutil
import tempfile
from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import datetime
from typing import Iterable, Optional, Union

import abutils
import polars as pl
from abutils import Sequence
from natsort import natsorted
from tqdm.auto import tqdm

from ..annotation.annotator import annotate
from ..assigners.mmseqs import MMseqs
from ..preprocess.merging import merge_fastqs

#  TODO: inputs/returns
#  --------------------
#
#  - inputs
#    - single Sequence (or something that can be processed by Sequence)
#    - Iterable of Sequence objects
#    - single FASTA/Q file
#    - directory of FASTA/Q files
#    - directory of FASTA/Q files with subdirectories (samples in subdirectories)
#
#  - returns
#    - Sequence objects
#      - should this happen automatically if an output directory is not provided?
#      - running via the API should allow outputs to be returned as Sequence objects or as files
#    - directory of AIRR/parquet files
#      - always return AIRR/parquet files when running at CLI -- maybe a check to ensure an output directory is provided?
#
#
#  the cleanest approach might be to just have two i/o arguments: 1) sequences and 2) project_path:
#
#  - sequences: Iterable[str, Sequence, Iterable]
#    - if str and is a file: process the file
#    - if str and is a directory: process the files in the directory
#    - if str and not either file or directory: assume str is a sequence, convert to Sequence and process
#    - if Sequence or Iterable[Sequence]: process accordingly
#
#  - project_path: Optional[str]
#    - if provided, str should be a directory path into which tmp, log and output files will be deposited
#    - if not provided:
#      - tmp files will be written to /tmp
#      - logs will not be generated
#      - outputs will be returned as Sequence objects


#  TODO: overall workflow
#  ----------------------
#
#  - set up directory structure
#    - user provides a "project path" and we make tmp, log and output (airr and/or parquet) directories within it
#
#  - preprocessing steps
#    - for now, just sequence merging for paired FASTQ inputsm but maybe primer/adapter trimming for FASTA inputs in the future?
#    - all samples are processed before proceeding to assignment
#    - results get deposited in project/merged
#
#  - VDJC assignment
#    - processes an entire input/merged file in a single job (thanks, MMseqs!)
#    - output is a single parquet file, deposited in project/tmp/assignment
#    - log failed assignments in project/logs/assignment as a single txt file per sample
#
#  - split the assignment output into job files
#    - deposited in project/tmp/assignment
#    - once the splitting is done, should call cleanup() on the assigner to remove tmp files that are no longer needed
#
#  - annotation
#    - annotation jobs run in parallel (multiprocessing)
#    - outputs are deposited in project/tmp/annotation
#    - log failed annotations (and successful annotations, if debug=True) as separate failed/succeeeded files in project/log/annotation
#    - temporary output and log files are collected into separate lists so they can be merged and then deleted.
#
#  - create outputs
#    - concat output files (using polars) and write to single tsv (AIRR) and/or parquet files in project/airr and project/parquet
#    - concatenate log files into single failed/succeeded text files in project/logs/annotation
#    - remove all of the temporary output and log files


def run(
    sequences: Union[str, Sequence, Iterable[Sequence]],
    project_path: Optional[str] = None,
    germline_database: str = "human",
    receptor: str = "bcr",
    output_format: Union[str, Iterable[str]] = "airr",
    umi_pattern: Optional[str] = None,
    umi_length: Optional[int] = None,
    merge: bool = False,
    merge_kwargs: Optional[dict] = None,
    interleaved_fastq: bool = False,
    chunksize: int = 500,
    mmseqs_chunksize: int = 1e6,
    mmseqs_threads: Optional[int] = None,
    n_processes: Optional[int] = None,
    copy_inputs_to_project: bool = False,
    verbose: bool = False,
    concise_logging: bool = False,
    started_from_cli: bool = False,
    debug: bool = False,
) -> Optional[Union[Iterable[Sequence], Sequence]]:
    """
    Annotate antibody or TCR sequences.

    \b
    command line arguments:
      SEQUENCES can be a FASTA/Q file or a directory of FASTA/Q files.
      PROJECT_PATH is the path to a directory in which tmp, log and output files will be deposited.

    \f

    Parameters
    ----------

    sequences : Union[str, Sequence, Iterable[Sequence]]
        The sequences to annotate. Can be one of the following:

          - ``str``: path to a FASTA/Q file, path to a directory of FASTA/Q files, or a single sequence, as a string
          - ``Sequence``: a single ``abutils.Sequence`` object
          - ``Iterable[Sequence]``: an iterable of ``abutils.Sequence`` objects

        .. note::
            If `sequences` is a directory path, files in the directory will be consumed recursively, including
            files in any subfolders.

    project_path : Optional[str] = None
        If provided, the path to a directory in which tmp, log and output files will be deposited. If not provided,
        annotated sequecnes will be returned as ``Sequence`` objects and log/tmp directories will be placed in
        ``"/tmp"``.

        .. warning::
            If ``debug=False`` (the default), the temporary directory will be removed during cleanup. If a
            directory named ``"tmp"`` exists in the provided `project_path`, it will be deleted.

    germline_database : str = "human",
        Name of the germline database to be used for assignment/annotation. Built-in options are
        "human", "mouse", and "macaque" and "humouse".

    receptor : str = "bcr",
        Name of the receptor to be used for assignment/annotation. Options are "bcr" and "tcr".

    output_format : Union[str, Iterable[str]] = "airr",
        Format of the output files. Options are "airr" and "parquet". If more than one output format is
        desired, a list of multiple output formats can be provided.

    umi_pattern : Optional[str], default=None
        Pattern to match for parsing the UMI sequence, or name of a built-in pattern.

    umi_length : Optional[int], default=None
        Length of the UMI sequence.

    merge : bool = False,
        Whether to merge FASTQ files prior to assignment. If ``True``, the ``merge_fastqs`` function will be used to merge
        FASTQ files.

    merge_kwargs : Optional[dict] = None,
        Keyword arguments to pass to the ``merge_fastqs`` function.

    interleaved_fastq : bool = False,
        Whether the input FASTQ files are interleaved. This is common when using data downloaded from the
        SRA_, which reuqires that paired FASTQ files be submitted as a single interleaved file rather than two
        separate files.

    chunksize : int = 500,
        Number of sequences to process at a time for annotation.

    mmseqs_chunksize : int = 1e6,
        Number of sequences to process at a time for MMseqs2 searches (VDJC assignment).

    n_processes : Optional[int] = None,
        Number of processes to use for annotation. If ``None``, the number of processes will be set to the number of
        available CPU cores.

    copy_inputs_to_project : bool = False,
        Whether to copy input sequences to the project directory.

    concise_logging : bool = False,
        Whether to use concise logging. If ``True``, the logging is more concise, more suitable for running
        abstar inside some other pipeline where only basic progress information is needed.

    verbose : bool = False,
        Whether to print verbose output.

    debug : bool = False,
        If ``True``, the following additional things will happen:
          - successfully annotated sequences will be logged in addition to sequences that errored during annotation
          - all tmp files will be retained
          - stdout and stderr from various third party tools (MMseqs, fastp, etc) will be captured and logged


    .. _SRA: https://www.ncbi.nlm.nih.gov/sra


    """
    # output format
    if isinstance(output_format, str):
        output_format = [output_format]

    # set up log/output/temp directories
    if project_path is not None:
        return_sequences = False
        project_path = os.path.abspath(project_path)
    else:
        return_sequences = True
        sequences_to_return = []
        output_format = ["parquet"]
        project_path = tempfile.TemporaryDirectory(prefix="abstar", dir="/tmp").name
    log_dir = os.path.join(project_path, "logs")
    abutils.io.make_dir(log_dir)
    temp_dir = os.path.join(project_path, "tmp")
    abutils.io.make_dir(temp_dir)
    for fmt in output_format:
        abutils.io.make_dir(os.path.join(project_path, fmt))

    # setup logging
    global logger
    if started_from_cli:
        logger = _setup_logging(
            log_dir,
            add_stream_handler=verbose,
            single_line_handler=True,
            debug=debug,
        )
    elif verbose or concise_logging:
        verbose = True
        logger = abutils.log.NotebookLogger(verbose=verbose, end="")
    else:
        logger = abutils.log.null_logger()

    if started_from_cli:
        _log_run_parameters(
            project_path=project_path,
            germline_database=germline_database,
            receptor=receptor,
            output_format=output_format,
            umi_pattern=umi_pattern,
            umi_length=umi_length,
            merge=merge,
            merge_kwargs=merge_kwargs,
            interleaved_fastq=interleaved_fastq,
            chunksize=chunksize,
            n_processes=n_processes,
            copy_inputs_to_project=copy_inputs_to_project,
            verbose=verbose,
            debug=debug,
        )

    # process input sequences
    sequence_files = _process_inputs(sequences, temp_dir)
    if copy_inputs_to_project:
        _copy_inputs_to_project(sequence_files, project_path)

    # merge FASTQ files
    if merge or interleaved_fastq:
        merge_dir = os.path.join(project_path, "merged")
        abutils.io.make_dir(merge_dir)
        merge_log_dir = os.path.join(log_dir, "merge_fastqs")
        abutils.io.make_dir(merge_log_dir)
        # log merge info
        if not concise_logging:
            logger.info("\n\n\n")
            logger.info("MERGE FASTQS\n")
            logger.info("============\n")
            logger.info(f"merge directory: {merge_dir}\n")
            # merging
        sequence_files = merge_fastqs(
            sequence_files,
            merge_dir,
            interleaved=interleaved_fastq,
            show_progress=verbose,
            log_directory=merge_log_dir,
            **merge_kwargs,
        )

    # print sequence file info
    if started_from_cli:
        _log_sequence_file_info(sequence_files)

    # annotation config
    if n_processes is None:
        n_processes = mp.cpu_count()
    annot_kwargs = {
        "output_directory": temp_dir,
        "germline_database": germline_database,
        "log_directory": log_dir,
        "umi_pattern": umi_pattern,
        "umi_length": umi_length,
        "debug": debug,
    }

    # initialize the assigner
    assigner = MMseqs(
        output_directory=temp_dir,
        log_directory=log_dir,
        germdb_name=germline_database,
        receptor=receptor,
        logger=logger,
        concise_logging=concise_logging,
        chunksize=mmseqs_chunksize,
        threads=mmseqs_threads,
        debug=debug,
    )

    # annotate sequences
    for sequence_file in natsorted(sequence_files):
        start_time = datetime.now()
        to_delete = []

        # parse sample name
        sample_name = ".".join(
            os.path.basename(sequence_file).rstrip(".gz").split(".")[:-1]
        )
        if not sample_name:  # if the input was Sequence object(s), not file(s)
            sample_name = "sequences"
        # log sample info
        if started_from_cli:
            logger.info("\n\n")
            logger.info("-" * (len(sample_name) + 4))
            logger.info("\n")
            logger.info(f"  {sample_name}\n")
            logger.info("-" * (len(sample_name) + 4))
            logger.info("\n")

        # assign VDJC genes, the returned assign_file is in parquet format
        assign_file, raw_sequence_count = assigner(sequence_file)
        assigner.cleanup()

        # split into annotation jobs
        split_assign_files = abutils.io.split_parquet(
            assign_file, temp_dir, num_rows=chunksize
        )

        # run annotation jobs
        annotated_files = []
        failed_log_files = []
        succeeded_log_files = []
        # log annotation info
        if concise_logging:
            logger.info("\nsequence annotation: ")
        else:
            logger.info("\n")
            logger.info("sequence annotation:\n")
        if verbose and started_from_cli:
            progress_bar = tqdm(
                total=len(split_assign_files),
                bar_format="{desc:<2.5}{percentage:3.0f}%|{bar:25}{r_bar}",
            )
        elif verbose:
            progress_bar = tqdm(
                total=len(split_assign_files),
            )
        with ProcessPoolExecutor(
            max_workers=n_processes,
            mp_context=mp.get_context("spawn"),
        ) as executor:
            futures = [
                executor.submit(annotate, f, **annot_kwargs) for f in split_assign_files
            ]
            for future in as_completed(futures):
                annotated, failed, succeeded = future.result()
                annotated_files.append(annotated)
                failed_log_files.append(failed)
                succeeded_log_files.append(succeeded)
                if verbose:
                    progress_bar.update(1)
        if verbose:
            progress_bar.close()

        # get output Sequences
        if return_sequences:
            annotated_sequences = abutils.io.read_parquet(annotated_files)
            sequences_to_return.extend(annotated_sequences)

            # log results summary
            sequence_count = len(annotated_sequences)
            duration = datetime.now() - start_time
            _log_results_summary(
                sequence_count=sequence_count,
                sequences_per_second=raw_sequence_count / duration.total_seconds(),
                seconds=duration.total_seconds(),
                concise_logging=concise_logging,
            )

        # or assemble output files (including logs)
        else:
            output_df = pl.scan_parquet(annotated_files)
            if "airr" in output_format:
                airr_file = os.path.join(project_path, f"airr/{sample_name}.tsv")
                output_df.sink_csv(airr_file, separator="\t")
            if "parquet" in output_format:
                parquet_file = os.path.join(
                    project_path, f"parquet/{sample_name}.parquet"
                )
                output_df.sink_parquet(parquet_file)
            # assemble logs
            failed_log_file = os.path.join(log_dir, f"{sample_name}.failed")
            _assemble_logs(failed_log_files, failed_log_file)
            if debug:
                # only log succeeded sequences if we're in debug mode
                succeeded_log_file = os.path.join(log_dir, f"{sample_name}.succeeded")
                _assemble_logs(succeeded_log_files, succeeded_log_file)

            # log results summary
            sequence_count = output_df.select(pl.count()).collect().row(0)[0]
            duration = datetime.now() - start_time
            _log_results_summary(
                sequence_count=sequence_count,
                sequences_per_second=raw_sequence_count / duration.total_seconds(),
                seconds=duration.total_seconds(),
                concise_logging=concise_logging,
            )

        # collect files for removal
        to_delete.append(assign_file)
        to_delete.extend(split_assign_files)
        to_delete.extend(annotated_files)
        to_delete.extend(succeeded_log_files)
        to_delete.extend(failed_log_files)

        # remove tmp files
        if not debug:
            _delete_files(to_delete)

    if return_sequences:
        if len(sequences_to_return) == 1:
            return sequences_to_return[0]
        return sequences_to_return
    else:
        return


def _process_inputs(
    sequences: Union[str, Sequence, Iterable[Sequence]],
    temp_dir: str,
) -> Iterable[str]:
    """
    Process the various inputs accepted by abstar and return a list of one or more sequence files.

    Parameters
    ----------
    sequences : Union[str, Sequence, Iterable[Sequence]]
        The sequences to process.

    temp_dir : str
        The path to a directory in which tmp files will be deposited.

    Returns
    -------
    sequence_files : Iterable[str]
        A list of one or more sequence files.
    """
    sequence_files = None
    if isinstance(sequences, str):
        # input is a string -- either a file/directory path or a raw sequence string
        if os.path.isfile(sequences):
            sequence_files = [os.path.abspath(sequences)]
        elif os.path.isdir(sequences):
            sequence_files = abutils.io.list_files(
                sequences,
                recursive=True,
                extension=["fasta", "fa", "fastq", "fq", "fasta.gz", "fastq.gz"],
            )
        else:
            sequences = Sequence(sequences)
    if isinstance(sequences, Sequence) or (
        isinstance(sequences, Iterable)
        and all(isinstance(s, Sequence) for s in sequences)
    ):
        # input is a Sequence or an iterable of Sequences
        if isinstance(sequences, Sequence):
            sequences = [sequences]
        temp_file = tempfile.NamedTemporaryFile(delete=False, dir=temp_dir, mode="w")
        fastas = [seq.fasta for seq in sequences]
        temp_file.write("\n".join(fastas))
        temp_file.close()
        sequence_files = [temp_file.name]
    if sequence_files is None:
        raise ValueError(
            "Invalid input sequences. Must be a path to a file or directory, a single sequence, or an iterable of sequences."
        )
    return natsorted(sequence_files)


def _copy_inputs_to_project(sequence_files: Iterable[str], project_path: str) -> None:
    """
    Copy input sequences to the project directory.

    Parameters
    ----------
    sequence_files : Iterable[str]
        A list of sequence files to copy.

    project_path : str
        The path to the project directory.
    """
    inputs_path = os.path.join(project_path, "input")
    abutils.io.make_dir(inputs_path)
    for f in sequence_files:
        shutil.copy(f, inputs_path)


def _assemble_logs(log_files: Iterable[str], combined_log_file: str) -> None:
    """
    Assemble log files into a single file.

    Parameters
    ----------
    log_files : Iterable[str]
        A list of log files to be assembled.

    combined_log_file : str
        The path to the file in which the combined log will be written.

    Returns
    -------
    None
    """
    with open(combined_log_file, "w") as f:
        for log_file in log_files:
            if log_file is not None:
                with open(log_file, "r") as log:
                    f.write(log.read())


def _delete_files(files: Iterable[str]) -> None:
    """
    Delete a list of files.
    """
    for f in files:
        if f is not None:
            if os.path.exists(f):
                os.remove(f)


# ===============================
#
#       PRINTING/LOGGING
#
# ===============================


def _setup_logging(
    log_dir: str,
    add_stream_handler: bool,
    single_line_handler: bool = False,
    debug: bool = False,
) -> None:
    logfile = os.path.join(log_dir, "abstar.log")
    abutils.log.setup_logging(
        logfile=logfile,
        add_stream_handler=add_stream_handler,
        single_line_handler=single_line_handler,
        debug=debug,
    )
    # global logger
    logger = abutils.log.get_logger(
        name="abstar",
        add_stream_handler=add_stream_handler,
        single_line_handler=single_line_handler,
    )
    return logger


def _log_run_parameters(
    project_path: str,
    germline_database: str,
    receptor: str,
    output_format: Union[str, Iterable[str]],
    umi_pattern: Optional[str],
    umi_length: Optional[int],
    merge: bool,
    merge_kwargs: Optional[dict],
    interleaved_fastq: bool,
    chunksize: int,
    n_processes: Optional[int],
    copy_inputs_to_project: bool,
    verbose: bool,
    debug: bool,
) -> None:
    logger.info("\n")
    # printing the splash line-by-line makes the log files look nicer
    for line in ABSTAR_SPLASH.split("\n"):
        logger.info(line + "\n")
    logger.info("\n")
    logger.info("RUN PARAMETERS\n")
    logger.info("===============\n")
    logger.info(f"PROJECT PATH: {project_path}\n")
    logger.info(f"GERMLINE DATABASE: {germline_database}\n")
    logger.info(f"RECEPTOR: {receptor}\n")
    logger.info(f"OUTPUT FORMAT: {', '.join(output_format)}\n")
    logger.info(f"UMI PATTERN: {umi_pattern}\n")
    logger.info(f"UMI LENGTH: {umi_length}\n")
    logger.info(f"MERGE: {merge}\n")
    logger.info(f"MERGE KWARGS: {merge_kwargs}\n")
    logger.info(f"INTERLEAVED FASTQ: {interleaved_fastq}\n")
    logger.info(f"CHUNKSIZE: {chunksize}\n")
    logger.info(
        f"NUM PROCESSES: {n_processes if n_processes is not None else 'auto'}\n"
    )
    logger.info(f"COPY INPUTS TO PROJECT: {copy_inputs_to_project}\n")
    logger.info(f"VERBOSE: {verbose}\n")
    logger.info(f"DEBUG: {debug}\n")


def _log_sequence_file_info(sequence_files: Iterable[str]) -> None:
    num_files = len(sequence_files)
    plural = "files" if num_files > 1 else "file"
    logger.info("\n\n\n")
    logger.info("INPUT FILES\n")
    logger.info("===========\n")
    logger.info(f"found {num_files} input {plural}:\n")
    if num_files < 6:
        for f in sequence_files:
            logger.info(f"  {os.path.basename(f)}\n")
    else:
        for f in sequence_files[:5]:
            logger.info(f"  {os.path.basename(f)}\n")
        logger.info(f"  ... and {num_files - 5} more\n")


def _log_results_summary(
    sequence_count: int,
    sequences_per_second: int,
    seconds: float,
    concise_logging: bool = False,
) -> None:
    if seconds < 60:
        hours = 0
        minutes = 0
        seconds = seconds
    elif seconds < 3600:
        hours = 0
        minutes = int(seconds / 60)
        seconds = int(seconds % 60)
    else:
        hours = int(seconds / 3600)
        minutes = (seconds % 3600) / 60
        seconds = seconds % 60
    duration_string = f"{hours:02}:{minutes:02}:{seconds:02.2f}"
    if concise_logging:
        logger.info(f"annotated sequences: {sequence_count:,}\n")
    else:
        logger.info("\n")
        logger.info(f"{sequence_count:,} sequences had an identifiable rearrangement\n")
    logger.info(
        f"time elapsed: {duration_string} ({sequences_per_second:,.2f} sequences/sec)\n"
    )


ABSTAR_SPLASH = """
         __         __ 
  ____ _/ /_  _____/ /_____ ______
 / __ `/ __ \/ ___/ __/ __ `/ ___/
/ /_/ / /_/ (__  ) /_/ /_/ / /    
\__,_/_.___/____/\__/\__,_/_/      
"""



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import pytest

from ..annotation.antibody import Antibody
from ..annotation.productivity import assess_productivity


@pytest.fixture
def productive_antibody():
    return Antibody(
        sequence="CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTAGCAGTTTTGGTATCAGCTGGGTGCGACAGGCCCCTGGGCAAGGGCTTGAGTGGCTGGGATGGAGCAGCACTGACAATGGTAACACAAACTATGCACAGAAGTTCCAGGGCAGAGTCACCATGACCACAGACACATCCACGACCACAGCCTACATGGAGCTGAGGAGCCTAAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGATTTAGGGCGGTGTACCAATACCGGGTGCTATCGCAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
        sequence_aa="QVQLVQSGAEVKKPGASVKVSCKASGYTFSSFGISWVRQAPGQGLEWLGWSSTDNGNTNYAQKFQGRVTMTTDTSTTTAYMELRSLRSDDTAVYYCARDLGRCTNTGCYRNWFDPWGQGTLVTVSS",
        v_call="IGHV1-18*01",
        j_call="IGHJ5*01",
        junction_aa="CARDLGRCTNTGCYRNWFDPW",
    )


@pytest.fixture
def stop_codon_antibody():
    return Antibody(
        sequence="CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTAGCAGTTTTGGTATCAGCTGGGTGCGACAGGCCCCTGGGCAAGGGCTTGAGTGGCTGGGATGGAGCAGCACTGACAATGGTAACACAAACTATGCACAGAAGTTCCAGGGCAGAGTCACCATGACCACAGACACATCCACGACCACAGCCTACATGGAGCTGAGGAGCCTAAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGATTTAGGGCGGTGTACCAATACCGGGTGCTATCGCAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
        sequence_aa="QVQLVQSGAEVKKPGASVKVSCKASGYTFSSFGISWVRQAPGQGLEWLGWSSTDNGNTNYAQKFQGRVT*TTDTSTTTAYMELRSLRSDDTAVYYCARDLGRCTNTGCYRNWFDPWGQGTLVTVSS",
        v_call="IGHV1-18*01",
        j_call="IGHJ5*01",
        junction_aa="CARDLGRCTNTGCYRNWFDPW",
    )


@pytest.fixture
def missing_conserved_cysteine_antibody():
    return Antibody(
        sequence="CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTAGCAGTTTTGGTATCAGCTGGGTGCGACAGGCCCCTGGGCAAGGGCTTGAGTGGCTGGGATGGAGCAGCACTGACAATGGTAACACAAACTATGCACAGAAGTTCCAGGGCAGAGTCACCATGACCACAGACACATCCACGACCACAGCCTACATGGAGCTGAGGAGCCTAAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGATTTAGGGCGGTGTACCAATACCGGGTGCTATCGCAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
        sequence_aa="QVQLVQSGAEVKKPGASVKVSCKASGYTFSSFGISWVRQAPGQGLEWLGWSSTDNGNTNYAQKFQGRVTMTTDTSTTTAYMELRSLRSDDTAVYYAARDLGRCTNTGCYRNWFDPWGQGTLVTVSS",
        v_call="IGHV1-18*01",
        j_call="IGHJ5*01",
        junction_aa="AARDLGRCTNTGCYRNWFDPW",
    )


@pytest.fixture
def missing_conserved_tryptophan_antibody():
    return Antibody(
        sequence="CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTAGCAGTTTTGGTATCAGCTGGGTGCGACAGGCCCCTGGGCAAGGGCTTGAGTGGCTGGGATGGAGCAGCACTGACAATGGTAACACAAACTATGCACAGAAGTTCCAGGGCAGAGTCACCATGACCACAGACACATCCACGACCACAGCCTACATGGAGCTGAGGAGCCTAAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGATTTAGGGCGGTGTACCAATACCGGGTGCTATCGCAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
        sequence_aa="QVQLVQSGAEVKKPGASVKVSCKASGYTFSSFGISWVRQAPGQGLEWLGWSSTDNGNTNYAQKFQGRVTMTTDTSTTTAYMELRSLRSDDTAVYYCARDLGRCTNTGCYRNWFDPAGQGTLVTVSS",
        v_call="IGHV1-18*01",
        j_call="IGHJ5*01",
        junction_aa="CARDLGRCTNTGCYRNWFDPA",
    )


@pytest.fixture
def locus_mismatch_antibody():
    return Antibody(
        sequence="CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTAGCAGTTTTGGTATCAGCTGGGTGCGACAGGCCCCTGGGCAAGGGCTTGAGTGGCTGGGATGGAGCAGCACTGACAATGGTAACACAAACTATGCACAGAAGTTCCAGGGCAGAGTCACCATGACCACAGACACATCCACGACCACAGCCTACATGGAGCTGAGGAGCCTAAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGATTTAGGGCGGTGTACCAATACCGGGTGCTATCGCAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
        sequence_aa="QVQLVQSGAEVKKPGASVKVSCKASGYTFSSFGISWVRQAPGQGLEWLGWSSTDNGNTNYAQKFQGRVTMTTDTSTTTAYMELRSLRSDDTAVYYCARDLGRCTNTGCYRNWFDPWGQGTLVTVSS",
        v_call="IGHV1-18*01",
        j_call="IGKJ5*01",
        junction_aa="CARDLGRCTNTGCYRNWFDPW",
    )


@pytest.fixture
def ambiguous_nucleotide_antibody():
    return Antibody(
        sequence="CAGGTTCAGCTGGTGCAGTCTGGAGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTAGCAGTTTTGGTATCAGCTGGGTGCGACAGGCCCCTGGGCAAGGGCNTGAGTGGCTGGGATGGAGCAGCACTGACAATGGTAACACAAACTATGCACAGAAGTTCCAGGGCAGAGTCACCATGACCACAGACACATCCACGACCACAGCCTACATGGAGCTGAGGAGCCTAAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGATTTAGGGCGGTGTACCAATACCGGGTGCTATCGCAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
        sequence_aa="QVQLVQSGAEVKKPGASVKVSCKASGYTFSSFGISWVRQAPGQGLEWLGWSSTDNGNTNYAQKFQGRVTMTTDTSTTTAYMELRSLRSDDTAVYYCARDLGRCTNTGCYRNWFDPWGQGTLVTVSS",
        v_call="IGHV1-18*01",
        j_call="IGHJ5*01",
        junction_aa="CARDLGRCTNTGCYRNWFDPW",
    )


def test_productive_antibody(productive_antibody):
    ab = assess_productivity(productive_antibody)
    assert ab.productive
    assert not ab.productivity_issues


def test_stop_codon_antibody(stop_codon_antibody):
    ab = assess_productivity(stop_codon_antibody)
    assert not ab.productive
    assert "stop codon" in ab.productivity_issues


def test_missing_conserved_cysteine_antibody(missing_conserved_cysteine_antibody):
    ab = assess_productivity(missing_conserved_cysteine_antibody)
    assert not ab.productive
    assert "conserved C" in ab.productivity_issues


def test_missing_conserved_tryptophan_antibody(missing_conserved_tryptophan_antibody):
    ab = assess_productivity(missing_conserved_tryptophan_antibody)
    assert not ab.productive
    assert "conserved W/F" in ab.productivity_issues


def test_locus_mismatch_antibody(locus_mismatch_antibody):
    ab = assess_productivity(locus_mismatch_antibody)
    assert not ab.productive
    assert "locus mismatch" in ab.productivity_issues


def test_ambiguous_nucleotide_antibody(ambiguous_nucleotide_antibody):
    ab = assess_productivity(ambiguous_nucleotide_antibody)
    assert not ab.productive
    assert "ambiguous nucleotide" in ab.productivity_issues



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import os
import traceback
from typing import Iterable, Optional, Union

import abutils

# import pandas as pd
import polars as pl

from .antibody import Antibody
from .germline import (
    get_germline,
    process_cgene_alignment,
    process_dgene_alignment,
    process_jgene_alignment,
    process_vgene_alignment,
    realign_germline,
    reassign_dgene,
)
from .indels import annotate_deletions, annotate_insertions
from .mutations import annotate_c_mutations, annotate_v_mutations
from .positions import get_gapped_sequence
from .productivity import assess_productivity
from .regions import get_region_sequence
from .schema import OUTPUT_SCHEMA
from .umi import parse_umis


def annotate(
    input_file: str,
    output_directory: str,
    germline_database: str,
    log_directory: Optional[str] = None,
    umi_pattern: Optional[str] = None,
    umi_length: Optional[int] = None,
    debug: bool = False,
) -> Iterable[Union[str, Optional[str], Optional[str]]]:
    """
    Annotates a Parquet file of V(D)J-assigned antibody sequences.

    Parameters
    ----------
    input_file : str
        Path to the input file, in Parquet format. The following fields are required:

            - sequence_id
            - sequence_input
            - quality
            - rev_comp
            - v_call
            - v_support
            - d_call
            - d_support
            - j_call
            - j_support
            - c_call
            - c_support

        although any of the values (aside from sequence_id, sequence_input, and rev_comp)
        can be None.

    output_directory : str
        Path to the directory where the annotated Parquet file will be written.

    germline_database : str
        Name of the germline database to use for annotation.

    log_directory : Optional[str], default=None
        Path to the directory where the log files will be written.

    umi_pattern : Optional[str], default=None
        Pattern to match for parsing the UMI sequence, or name of a built-in pattern.

    umi_length : Optional[int], default=None
        Length of the UMI sequence.

    debug : bool, default=False
        Whether to write log files for failed and succeeded sequences.

    Returns
    -------
    output_file : str
        Path to the output Parquet file containing annotated sequences.

    failed_logfile : Optional[str]
        Path to the log file for failed sequences. Only returned if
        `log_directory` is provided

    succeeded_logfile : Optional[str]
        Path to the log file for succeeded sequences. Only returned if
        `debug=True` and `log_directory` is provided.

    """
    # load the input parquet file
    df = pl.read_parquet(input_file)

    # do annotations
    annotated = []
    for r in df.iter_rows(named=True):
        ab = Antibody(**r)
        try:
            ab = annotate_single_sequence(
                ab=ab,
                germline_database=germline_database,
                umi_pattern=umi_pattern,
                umi_length=umi_length,
                debug=debug,
            )
        except Exception as e:
            ab.exception("ANNOTATION EXCEPTION", traceback.format_exc())
        annotated.append(ab)

    # gather the results
    failed = [a for a in annotated if a.exceptions]
    succeeded = [a for a in annotated if not a.exceptions]
    succeeded_df = pl.DataFrame([s.to_dict() for s in succeeded], schema=OUTPUT_SCHEMA)

    # write logs
    basename = os.path.basename(input_file)
    if log_directory:
        failed_logfile = os.path.join(log_directory, f"{basename}.failed")
        with open(failed_logfile, "w") as f:
            for fail in failed:
                f.write(fail.format_log())
        if debug:
            succeeded_logfile = os.path.join(log_directory, f"{basename}.succeeded")
            with open(succeeded_logfile, "w") as f:
                for succ in succeeded:
                    f.write(succ.format_log())

    # write outputs
    output_file = os.path.join(output_directory, f"{basename}_annotated.parquet")
    succeeded_df.write_parquet(output_file)

    # returns
    if log_directory:
        if debug:
            return output_file, failed_logfile, succeeded_logfile
        else:
            return output_file, failed_logfile, None
    else:
        return output_file, None, None


def annotate_single_sequence(
    ab: Antibody,
    germline_database: str,
    umi_pattern: Optional[str] = None,
    umi_length: Optional[int] = None,
    debug: bool = False,
):
    # log information from the assigner
    ab.log("=" * (len(str(ab.sequence_id)) + 15))
    ab.log(" SEQUENCE ID:", ab.sequence_id)
    ab.log("=" * (len(str(ab.sequence_id)) + 15) + "\n")
    ab.log(f">{ab.sequence_id}\n{ab.sequence_input}\n")
    ab.log(f"{ab.v_call} | {ab.d_call} | {ab.j_call}")
    ab.log("REV COMP:", ab.rev_comp)

    # germline database, locus, and receptor type
    ab.germline_database = germline_database
    ab.locus = ab.v_call[:3].upper()
    ab.receptor_type = "bcr" if ab.locus[:2] == "IG" else "tcr"
    ab.log("GERMLINE DATABASE:", ab.germline_database)
    ab.log("LOCUS:", ab.locus)

    # umi
    if any([umi_pattern is not None, umi_length is not None]):
        ab.umi = parse_umis(
            ab.sequence_input,
            pattern=umi_pattern,
            length=umi_length,
        )

    # orient the input sequence
    if ab.rev_comp:
        ab.sequence_oriented = abutils.tl.reverse_complement(ab.sequence_input)
    else:
        ab.sequence_oriented = ab.sequence_input
    ab.log("SEQUENCE ORIENTED:", ab.sequence_oriented)

    # check germline calls for species name
    # (mainly for mixed species databases like humouse)
    if len(vsplit := ab.v_call.split("__")) > 1:
        ab.v_call, ab.species = vsplit
    else:
        ab.species = ab.germline_database
    if len(jsplit := ab.j_call.split("__")) > 1:
        ab.j_call, _ = jsplit
    if ab.d_call is not None:
        if len(dsplit := ab.d_call.split("__")) > 1:
            ab.d_call, _ = dsplit
    if ab.c_call is not None:
        if len(c_split := ab.c_call.split("__")) > 1:
            ab.c_call, _ = c_split
    ab.log("SPECIES:", ab.species)

    # get genes from calls
    ab.v_gene = ab.v_call.split("*")[0]
    ab.j_gene = ab.j_call.split("*")[0]
    if ab.d_call is not None:
        ab.d_gene = ab.d_call.split("*")[0]
    if ab.c_call is not None:
        ab.c_gene = ab.c_call.split("*")[0]

    ab.log("\n--------")
    ab.log(" V GENE")
    ab.log("--------\n")

    # v-gene realignment
    v_sg, v_loc = realign_germline(
        sequence=ab.sequence_oriented,
        germline_name=ab.v_call,
        germdb_name=ab.germline_database,
        imgt_gapped=False,
        semiglobal_aln_params=ALIGNMENT_PARAMS,
        local_aln_params=ALIGNMENT_PARAMS,
    )
    ab.log("V GERMLINE:", v_sg.query.sequence)

    ab.log("SEMIGLOBAL ALIGNMENT:")
    ab.log(f"     QUERY: {v_sg.aligned_query}")
    ab.log(f"            {v_sg.alignment_midline}")
    ab.log(f"  GERMLINE: {v_sg.aligned_target}")
    ab.log("LOCAL ALIGNMENT:")
    ab.log(f"     QUERY: {v_loc.aligned_query}")
    ab.log(f"            {v_loc.alignment_midline}")
    ab.log(f"  GERMLINE: {v_loc.aligned_target}")

    ab = process_vgene_alignment(semiglobal_aln=v_sg, local_aln=v_loc, ab=ab)
    ab.log("SEMIGLOBAL QUERY START:", v_sg.query_begin)
    ab.log("SEMIGLOBAL GERMLINE START:", v_sg.target_begin)
    ab.log("LOCAL QUERY START:", v_loc.query_begin)
    ab.log("LOCAL GERMLINE START:", v_loc.target_begin)
    ab.log("V SEQUENCE START:", ab.v_sequence_start)
    ab.log("V SEQUENCE END:", ab.v_sequence_end)
    ab.log("V GERMLINE START:", ab.v_germline_start)
    ab.log("V GERMLINE END:", ab.v_germline_end)
    ab.log("V FRAME:", ab.frame)
    ab.log("V SEQUENCE:", ab.v_sequence)
    ab.log("V GERMLINE:", ab.v_germline)
    ab.log("V SEQUENCE AA:", ab.v_sequence_aa)
    ab.log("V GERMLINE AA:", ab.v_germline_aa)

    # global alignment of sequence and germline
    # using the start/end positions determined parsed from the
    # semiglobal and local alignments
    v_global = abutils.tl.global_alignment(
        ab.v_sequence,
        ab.v_germline,
        **ALIGNMENT_PARAMS,
    )
    ab.log("GLOBAL ALIGNMENT:")
    ab.log(f"     QUERY: {v_global.aligned_query}")
    ab.log(f"            {v_global.alignment_midline}")
    ab.log(f"  GERMLINE: {v_global.aligned_target}")

    # global alignment of the AA sequences
    v_global_aa = abutils.tl.global_alignment(
        ab.v_sequence_aa,
        ab.v_germline_aa,
        **ALIGNMENT_PARAMS,
    )
    ab.log("GLOBAL ALIGNMENT AA:")
    ab.log(f"     QUERY: {v_global_aa.aligned_query}")
    ab.log(f"            {v_global_aa.alignment_midline}")
    ab.log(f"  GERMLINE: {v_global_aa.aligned_target}")

    # gapped V-gene germline
    ab.v_germline_gapped = get_germline(
        ab.v_call, ab.germline_database, imgt_gapped=True, exact_match=True
    ).sequence
    ab.v_germline_gapped_aa = abutils.tl.translate(
        ab.v_germline_gapped, allow_dots=True
    )
    ab.log("GAPPED GERMLINE:", ab.v_germline_gapped)
    ab.log("GAPPED GERMLINE AA:", ab.v_germline_gapped_aa)
    # ab.v_germline_gapped = ab.v_germline_gapped
    # ab.v_germline_gapped_aa = ab.v_germline_gapped_aa

    # gapped V-gene sequence
    ab.v_sequence_gapped = get_gapped_sequence(
        aligned_sequence=v_global.aligned_query,
        aligned_germline=v_global.aligned_target,
        gapped_germline=ab.v_germline_gapped,
        germline_start=ab.v_germline_start,
    )
    ab.v_sequence_gapped_aa = abutils.tl.translate(
        ab.v_sequence_gapped, frame=ab.frame, allow_dots=True
    )
    ab.log("GAPPED SEQUENCE:", ab.v_sequence_gapped)
    ab.log("GAPPED SEQUENCE AA:", ab.v_sequence_gapped_aa)
    # ab.v_sequence_gapped = ab.v_sequence_gapped
    # ab.v_sequence_gapped_aa = ab.v_sequence_gapped_aa

    # insertions
    if "-" in v_loc.aligned_target:
        ab.v_insertions = annotate_insertions(
            aligned_sequence=v_global.aligned_query,
            aligned_germline=v_global.aligned_target,
            gapped_germline=ab.v_germline_gapped,
            germline_start=ab.v_germline_start,
        )
        ab.log("V INSERTIONS:", ab.v_insertions)
        insertions = ab.v_insertions.split("|")
        cumulative_ins_length = sum(
            [int(i.split(">")[0].split(":")[1]) for i in insertions]
        )
    else:
        cumulative_ins_length = 0

    # deletions
    if "-" in v_loc.aligned_query:
        ab.v_deletions = annotate_deletions(
            aligned_sequence=v_loc.aligned_query,
            aligned_germline=v_loc.aligned_target,
            gapped_germline=ab.v_germline_gapped,
            germline_start=ab.v_germline_start,
        )
        ab.log("V DELETIONS:", ab.v_deletions)
        deletions = ab.v_deletions.split("|")
        cumulative_del_length = sum(
            [int(d.split(">")[0].split(":")[1]) for d in deletions]
        )
    else:
        cumulative_del_length = 0

    # only raise a productivity issue if the combination of all indels causes a frameshift
    # this avoids incorrectly flagging as non-productive a sequence with compensatory out-of-frame indels
    if (cumulative_ins_length - cumulative_del_length) % 3:
        ab.productivity_issues.append("out-of-frame indel(s)")
        ab.v_frameshift = True

    ab.log("\n--------")
    ab.log(" J GENE")
    ab.log("--------\n")

    # j-gene realignment
    jquery = ab.sequence_oriented[ab.v_sequence_end :]
    if len(jquery) == 0:
        raise ValueError(
            f"Query sequence for J-gene realignment is empty for {ab.sequence_id}"
        )
    j_sg, j_loc = realign_germline(
        sequence=jquery,
        germline_name=ab.j_call,
        germdb_name=ab.germline_database,
        imgt_gapped=False,
        semiglobal_aln_params=ALIGNMENT_PARAMS,
        local_aln_params=ALIGNMENT_PARAMS,
    )
    ab.log("SEMIGLOBAL ALIGNMENT:")
    ab.log(f"     QUERY: {j_sg.aligned_query}")
    ab.log(f"            {j_sg.alignment_midline}")
    ab.log(f"  GERMLINE: {j_sg.aligned_target}")
    ab.log("LOCAL ALIGNMENT:")
    ab.log(f"     QUERY: {j_loc.aligned_query}")
    ab.log(f"            {j_loc.alignment_midline}")
    ab.log(f"  GERMLINE: {j_loc.aligned_target}")

    ab = process_jgene_alignment(
        oriented_input=ab.sequence_oriented,
        v_sequence_end=ab.v_sequence_end,
        semiglobal_aln=j_sg,
        local_aln=j_loc,
        ab=ab,
    )
    ab.log("SEMIGLOBAL QUERY START:", j_sg.query_begin)
    ab.log("LOCAL QUERY START:", j_loc.query_begin)
    ab.log("J SEQUENCE START:", ab.j_sequence_start)
    ab.log("J SEQUENCE END:", ab.j_sequence_end)
    ab.log("J GERMLINE START:", ab.j_germline_start)
    ab.log("J GERMLINE END:", ab.j_germline_end)
    ab.log("J SEQUENCE:", ab.j_sequence)
    ab.log("J GERMLINE:", ab.j_germline)

    ab.log("\n--------")
    ab.log(" D GENE")
    ab.log("--------\n")

    # d-gene realignment
    dquery = ab.sequence_oriented[ab.v_sequence_end : ab.j_sequence_start]
    # if the assigner made a d-gene call
    if dquery and ab.d_call is not None:
        _, d_loc = realign_germline(
            sequence=dquery,
            germline_name=ab.d_call,
            germdb_name=ab.germline_database,
            imgt_gapped=False,
            skip_semiglobal=True,
            local_aln_params=ALIGNMENT_PARAMS,
        )
    # if not, we can try again using local pairwise alignment (for IGH/TRA/TRD chains only)
    elif len(dquery) >= 5 and ab.locus in [
        "IGH",
        "TRA",
        "TRD",
    ]:  # TODO: make the minimum d-gene alignment length a user-controllable parameter?
        d_loc = reassign_dgene(
            sequence=dquery,
            germdb_name=ab.germline_database,
            aln_params=ALIGNMENT_PARAMS,
        )
    # maybe there's not a d-gene (or it's a light/TRB/TRG chain)
    else:
        ab.d_call = None
        ab.d_gene = None
        ab.d_score = None
        d_loc = None
    # process the d-gene alignment
    if d_loc is not None:
        ab = process_dgene_alignment(
            oriented_input=ab.sequence_oriented,
            v_sequence_end=ab.v_sequence_end,
            local_aln=d_loc,
            ab=ab,
        )
        ab.np1 = ab.sequence_oriented[ab.v_sequence_end : ab.d_sequence_start]
        ab.np2 = ab.sequence_oriented[ab.d_sequence_end : ab.j_sequence_start]
        ab.np1_length = len(ab.np1)
        ab.np2_length = len(ab.np2)
    else:
        ab.np1 = ab.sequence_oriented[ab.v_sequence_end : ab.j_sequence_start]
        ab.np1_length = len(ab.np1)

    if d_loc is not None:
        ab.log("LOCAL ALIGNMENT:")
        ab.log(f"     QUERY: {d_loc.aligned_query}")
        ab.log(f"            {d_loc.alignment_midline}")
        ab.log(f"  GERMLINE: {d_loc.aligned_target}")

        ab.log("LOCAL QUERY START:", d_loc.query_begin)
        ab.log("D SEQUENCE START:", ab.d_sequence_start)
        ab.log("D SEQUENCE END:", ab.d_sequence_end)
        ab.log("D GERMLINE START:", ab.d_germline_start)
        ab.log("D GERMLINE END:", ab.d_germline_end)
        ab.log("D FRAME:", ab.d_frame)
        ab.log("D SEQUENCE:", ab.d_sequence)
        ab.log("D GERMLINE:", ab.d_germline)
        ab.log("NP1 SEQUENCE:", ab.np1)
        ab.log("NP1 LENGTH:", ab.np1_length)
        ab.log("NP2 SEQUENCE:", ab.np2)
        ab.log("NP2 LENGTH:", ab.np2_length)

    ab.log("\n----------------")
    ab.log(" CONSTANT REGION")
    ab.log("----------------\n")

    # constant region realignment
    cquery = ab.sequence_oriented[ab.j_sequence_end :]
    # if the assigner made a c-gene call
    if cquery and ab.c_call is not None:
        c_sg, c_loc = realign_germline(
            sequence=cquery,
            germline_name=ab.c_call,
            germdb_name=ab.germline_database,
            imgt_gapped=False,
            semiglobal_aln_params=ALIGNMENT_PARAMS,
            local_aln_params=ALIGNMENT_PARAMS,
            force_constant=True,
        )

        ab.log("SEMIGLOBAL ALIGNMENT:")
        ab.log(f"     QUERY: {c_sg.aligned_query}")
        ab.log(f"            {c_sg.alignment_midline}")
        ab.log(f"  GERMLINE: {c_sg.aligned_target}")
        ab.log("LOCAL ALIGNMENT:")
        ab.log(f"     QUERY: {c_loc.aligned_query}")
        ab.log(f"            {c_loc.alignment_midline}")
        ab.log(f"  GERMLINE: {c_loc.aligned_target}")

        ab = process_cgene_alignment(
            oriented_input=ab.sequence_oriented,
            j_sequence_end=ab.j_sequence_end,
            semiglobal_aln=c_sg,
            local_aln=c_loc,
            ab=ab,
        )
        ab.log("SEMIGLOBAL QUERY START:", c_sg.query_begin)
        ab.log("LOCAL QUERY START:", c_loc.query_begin)
        ab.log("C SEQUENCE START:", ab.c_sequence_start)
        ab.log("C SEQUENCE END:", ab.c_sequence_end)
        ab.log("C GERMLINE START:", ab.c_germline_start)
        ab.log("C GERMLINE END:", ab.c_germline_end)
        ab.log("C SEQUENCE:", ab.c_sequence)
        ab.log("C GERMLINE:", ab.c_germline)
        ab.log("C SEQUENCE AA:", ab.c_sequence_aa)
        ab.log("C GERMLINE AA:", ab.c_germline_aa)

        # global nucleotide alignment
        c_global = abutils.tl.global_alignment(
            ab.c_sequence,
            ab.c_germline,
            **ALIGNMENT_PARAMS,
        )
        ab.log("GLOBAL ALIGNMENT:")
        ab.log(f"     QUERY: {c_global.aligned_query}")
        ab.log(f"            {c_global.alignment_midline}")
        ab.log(f"  GERMLINE: {c_global.aligned_target}")

        # # verify that the AA sequences are not empty
        # if len(ab.v_sequence_aa) == 0:
        #     raise ValueError(f"V-gene AA sequence is empty for {ab.sequence_id}")
        # if len(ab.v_germline_aa) == 0:
        #     raise ValueError(f"V-gene germline AA sequence is empty for {ab.sequence_id}")

        # global alignment of the AA sequences
        c_global_aa = abutils.tl.global_alignment(
            ab.c_sequence_aa,
            ab.c_germline_aa,
            **ALIGNMENT_PARAMS,
        )
        ab.log("GLOBAL ALIGNMENT AA:")
        ab.log(f"     QUERY: {c_global_aa.aligned_query}")
        ab.log(f"            {c_global_aa.alignment_midline}")
        ab.log(f"  GERMLINE: {c_global_aa.aligned_target}")

        # gapped Constant region germline
        ab.c_germline_gapped = get_germline(
            ab.c_call,
            ab.germline_database,
            imgt_gapped=True,
            exact_match=True,
            force_constant=True,
        ).sequence
        ab.c_germline_gapped_aa = abutils.tl.translate(
            ab.c_germline_gapped, allow_dots=True
        )
        ab.log("GAPPED GERMLINE:", ab.c_germline_gapped)
        ab.log("GAPPED GERMLINE AA:", ab.c_germline_gapped_aa)
        # ab.v_germline_gapped = ab.v_germline_gapped
        # ab.v_germline_gapped_aa = ab.v_germline_gapped_aa

        # gapped Constant region sequence
        ab.c_sequence_gapped = get_gapped_sequence(
            aligned_sequence=c_global.aligned_query,
            aligned_germline=c_global.aligned_target,
            gapped_germline=ab.c_germline_gapped,
            germline_start=ab.c_germline_start,
        )
        ab.c_sequence_gapped_aa = abutils.tl.translate(
            ab.c_sequence_gapped, frame=ab.frame, allow_dots=True
        )

        # nucleotide mutations
        complete_c_germline = c_sg.target.sequence
        ab = annotate_c_mutations(
            aligned_sequence=c_global.aligned_query,
            aligned_germline=c_global.aligned_target,
            gapped_germline=complete_c_germline,
            germline_start=ab.c_germline_start,
            is_aa=False,
            ab=ab,
        )
        ab.log("CONSTANT REGION MUTATIONS:", ab.c_mutations)
        ab.log("CONSTANT REGION MUTATION COUNT:", ab.c_mutation_count)

        # amino acid mutations
        complete_c_germline_aa = abutils.tl.translate(complete_c_germline)
        ab = annotate_c_mutations(
            aligned_sequence=c_global_aa.aligned_query,
            aligned_germline=c_global_aa.aligned_target,
            gapped_germline=complete_c_germline_aa,
            germline_start=ab.c_germline_start // 3,
            is_aa=True,
            ab=ab,
        )
        ab.log("CONSTANT REGION MUTATIONS AA:", ab.c_mutations_aa)
        ab.log("CONSTANT REGION MUTATION COUNT AA:", ab.c_mutation_count_aa)

        # calculate identify (nt and aa)
        ab.c_identity = 1 - ab.c_mutation_count / len(ab.c_germline)
        ab.c_identity_aa = 1 - ab.c_mutation_count_aa / len(ab.c_germline_aa)
        ab.log("CONSTANT REGION IDENTITY:", ab.c_identity)
        ab.log("CONSTANT REGION IDENTITY AA:", ab.c_identity_aa)

        # insertions
        if "-" in c_loc.aligned_target:
            ab.c_insertions = annotate_insertions(
                aligned_sequence=c_global.aligned_query,
                aligned_germline=c_global.aligned_target,
                gapped_germline=complete_c_germline,
                germline_start=ab.c_germline_start,
            )
            ab.log("CONSTANT REGION INSERTIONS:", ab.c_insertions)

        # deletions
        if "-" in c_loc.aligned_query:
            ab.c_deletions = annotate_deletions(
                aligned_sequence=c_loc.aligned_query,
                aligned_germline=c_loc.aligned_target,
                gapped_germline=complete_c_germline,
                germline_start=ab.c_germline_start,
            )
            ab.log("CONSTANT REGION DELETIONS:", ab.c_deletions)

    ab.log("\n--------------")
    ab.log(" VDJ ASSEMBLY")
    ab.log("--------------\n")

    # assemble the full V(D)J and germline sequences
    ab.sequence = ab.v_sequence + ab.np1
    ab.germline = ab.v_germline + ab.np1
    ab.sequence_gapped = ab.v_sequence_gapped + ab.np1
    ab.germline_gapped = ab.v_germline_gapped + ab.np1
    if ab.d_sequence is not None:
        ab.sequence += ab.d_sequence
        ab.germline += ab.d_germline
        ab.sequence_gapped += ab.d_sequence
        ab.germline_gapped += ab.d_germline
    if ab.np2 is not None:
        ab.sequence += ab.np2
        ab.germline += ab.np2
        ab.sequence_gapped += ab.np2
        ab.germline_gapped += ab.np2
    ab.sequence += ab.j_sequence
    ab.germline += ab.j_germline
    ab.sequence_gapped += ab.j_sequence
    ab.germline_gapped += ab.j_germline
    if ab.c_sequence is not None:
        # only compute VDJC sequences if there is a constant region
        ab.sequence_vdjc = ab.sequence + ab.c_sequence
        ab.germline_vdjc = ab.germline + ab.c_germline
        ab.sequence_vdjc_gapped = ab.sequence_gapped + ab.c_sequence
        ab.germline_vdjc_gapped = ab.germline_gapped + ab.c_germline

    ab.log("SEQUENCE:", ab.sequence)
    ab.log("GERMLINE:", ab.germline)
    ab.log("GAPPED SEQUENCE:", ab.sequence_gapped)
    ab.log("GAPPED GERMLINE:", ab.germline_gapped)

    ab.log("VDJC SEQUENCE:", ab.sequence_vdjc)
    ab.log("VDJC GERMLINE:", ab.germline_vdjc)
    ab.log("GAPPED VDJC SEQUENCE:", ab.sequence_vdjc_gapped)
    ab.log("GAPPED VDJC GERMLINE:", ab.germline_vdjc_gapped)

    # translated sequences
    ab.sequence_aa = abutils.tl.translate(ab.sequence, frame=ab.frame)
    ab.germline_aa = abutils.tl.translate(ab.germline, frame=ab.frame)
    # ab.sequence_gapped_aa = abutils.tl.translate(
    #     ab.sequence_gapped, frame=ab.frame, allow_dots=True
    # )
    # ab.germline_gapped_aa = abutils.tl.translate(
    #     ab.germline_gapped, frame=ab.frame, allow_dots=True
    # )

    if ab.c_sequence is not None:
        # only compute gapped VDJC sequences if there is a constant region
        ab.sequence_vdjc_aa = abutils.tl.translate(ab.sequence_vdjc, frame=ab.frame)
        ab.germline_vdjc_aa = abutils.tl.translate(ab.germline_vdjc, frame=ab.frame)
        # ab.sequence_vdjc_gapped_aa = abutils.tl.translate(
        #     ab.sequence_vdjc_gapped, frame=ab.frame, allow_dots=True
        # )
        # ab.germline_vdjc_gapped_aa = abutils.tl.translate(
        #     ab.germline_vdjc_gapped, frame=ab.frame, allow_dots=True
        # )
    ab.log("SEQUENCE AA:", ab.sequence_aa)
    ab.log("GERMLINE AA:", ab.germline_aa)
    # ab.log("GAPPED SEQUENCE AA:", ab.sequence_gapped_aa)
    # ab.log("GAPPED GERMLINE AA:", ab.germline_gapped_aa)

    ab.log("VDJC SEQUENCE AA:", ab.sequence_vdjc_aa)
    ab.log("VDJC GERMLINE AA:", ab.germline_vdjc_aa)
    # ab.log("GAPPED VDJC SEQUENCE AA:", ab.sequence_vdjc_gapped_aa)
    # ab.log("GAPPED VDJC GERMLINE AA:", ab.germline_vdjc_gapped_aa)

    # align assembled V(D)J and germline nucleotide sequences
    nt_aln = abutils.tl.global_alignment(
        query=ab.sequence,
        target=ab.germline,
        **ALIGNMENT_PARAMS,
    )
    ab.sequence_alignment = nt_aln.aligned_query
    ab.germline_alignment = nt_aln.aligned_target
    ab.log(
        "SEQUENCE ALIGNMENT:",
        f"     QUERY: {nt_aln.aligned_query}",
        f"            {nt_aln.alignment_midline}",
        f"  GERMLINE: {nt_aln.aligned_target}",
        separator="\n",
    )

    # align assembled V(D)J and germline amino acid sequences
    aa_aln = abutils.tl.global_alignment(
        query=ab.sequence_aa,
        target=ab.germline_aa,
        **ALIGNMENT_PARAMS,
    )
    ab.sequence_alignment_aa = aa_aln.aligned_query
    ab.germline_alignment_aa = aa_aln.aligned_target
    ab.log(
        "SEQUENCE AA ALIGNMENT:",
        f"     QUERY: {aa_aln.aligned_query}",
        f"            {aa_aln.alignment_midline}",
        f"  GERMLINE: {aa_aln.aligned_target}",
        separator="\n",
    )

    # check for complete VDJ
    if ab.v_germline_start == 0 and ab.j_germline_end == len(j_sg.target.sequence):
        ab.complete_vdj = True
    ab.log("COMPLETE VDJ:", ab.complete_vdj)

    ab.log("\n----------")
    ab.log(" JUNCTION")
    ab.log("----------\n")

    # junction start
    germ_fr3_sequence = ab.v_germline_gapped[196:309].replace(".", "")
    fr3_sg = abutils.tl.semiglobal_alignment(
        query=ab.sequence_oriented,
        target=germ_fr3_sequence,
        **ALIGNMENT_PARAMS,
    )
    ab.junction_start = fr3_sg.query_end + 1

    # junction end
    if ab.locus in ["IGH", "TRA", "TRD"]:
        germ_fr4_sequence = j_sg.target[-34:]
    else:
        germ_fr4_sequence = j_sg.target[-31:]
    fr4_sg = abutils.tl.semiglobal_alignment(
        query=ab.sequence_oriented,
        target=germ_fr4_sequence,
        **ALIGNMENT_PARAMS,
    )
    ab.junction_end = fr4_sg.query_begin + 3  # germ_fr4_sequence includes the W/F

    # junction sequence
    ab.junction = ab.sequence_oriented[ab.junction_start : ab.junction_end]
    ab.junction_aa = abutils.tl.translate(ab.junction)
    ab.log("JUNCTION:", ab.junction)
    ab.log("JUNCTION AA:", ab.junction_aa)

    # CDR3 sequence and length
    ab.cdr3 = ab.junction[3:-3]
    ab.cdr3_aa = ab.junction_aa[1:-1]
    ab.cdr3_length = len(ab.cdr3_aa)
    ab.log("CDR3:", ab.cdr3)
    ab.log("CDR3 AA:", ab.cdr3_aa)
    ab.log("CDR3 LENGTH:", ab.cdr3_length)

    ab.log("\n-----------")
    ab.log(" MUTATIONS")
    ab.log("-----------\n")

    # nucleotide mutations
    ab = annotate_v_mutations(
        aligned_sequence=v_global.aligned_query,
        aligned_germline=v_global.aligned_target,
        gapped_germline=ab.v_germline_gapped,
        germline_start=ab.v_germline_start,
        is_aa=False,
        ab=ab,
        debug=debug,
    )
    ab.log("V MUTATIONS:", ab.v_mutations)
    ab.log("V MUTATION COUNT:", ab.v_mutation_count)

    # amino acid mutations
    ab = annotate_v_mutations(
        aligned_sequence=v_global_aa.aligned_query,
        aligned_germline=v_global_aa.aligned_target,
        gapped_germline=ab.v_germline_gapped_aa,
        germline_start=ab.v_germline_start // 3,
        is_aa=True,
        ab=ab,
        debug=debug,
    )
    ab.log("V MUTATIONS AA:", ab.v_mutations_aa)
    ab.log("V MUTATION COUNT AA:", ab.v_mutation_count_aa)

    # calculate identify (nt and aa)
    ab.v_identity = 1 - ab.v_mutation_count / len(ab.v_germline)
    ab.v_identity_aa = 1 - ab.v_mutation_count_aa / len(ab.v_germline_aa)
    ab.log("V IDENTITY:", ab.v_identity)
    ab.log("V IDENTITY AA:", ab.v_identity_aa)

    ab.log("\n---------")
    ab.log(" REGIONS")
    ab.log("---------\n")

    # V regions
    v_regions = ["fwr1", "cdr1", "fwr2", "cdr2", "fwr3"]
    ab.log("ALIGNED SEQUENCE:", v_global.aligned_query)
    ab.log("ALIGNED GERMLINE:", v_global.aligned_target)
    ab.log("GAPPED GERMLINE:", ab.v_germline_gapped)
    ab.log("GERMLINE START:", ab.v_germline_start + 1)
    ab.log("ALIGNED QUERY AA:", v_global_aa.aligned_query)
    ab.log("ALIGNED GERMLINE AA:", v_global_aa.aligned_target)
    ab.log("GAPPED GERMLINE AA:", ab.v_germline_gapped_aa)
    ab.log("GERMLINE START AA:", ab.v_germline_start // 3 + 1)

    for region in v_regions:
        # nucleotide region
        region_sequence = get_region_sequence(
            region,
            aligned_sequence=v_global.aligned_query,
            aligned_germline=v_global.aligned_target,
            gapped_germline=ab.v_germline_gapped,
            ab=ab,
            germline_start=ab.v_germline_start + 1,  # needs to be 1-indexed
        )
        setattr(ab, f"{region}", region_sequence)
        ab.log(f"{region.upper()} SEQUENCE:", region_sequence)
        # amino acid region
        region_sequence_aa = get_region_sequence(
            region,
            aligned_sequence=v_global_aa.aligned_query,
            aligned_germline=v_global_aa.aligned_target,
            gapped_germline=ab.v_germline_gapped_aa,
            ab=ab,
            germline_start=ab.v_germline_start // 3 + 1,  # needs to be 1-indexed
            aa=True,
        )
        setattr(ab, f"{region}_aa", region_sequence_aa)
        ab.log(f"{region.upper()} SEQUENCE AA:", region_sequence_aa)

    # J regions
    fwr4_start = fr4_sg.query_begin
    fwr4_end = fr4_sg.query_end
    ab.fwr4 = fr4_sg.aligned_query[fwr4_start:fwr4_end]
    ab.fwr4_aa = abutils.tl.translate(ab.fwr4)
    ab.log("FR4 SEQUENCE:", ab.fwr4)
    ab.log("FR4 SEQUENCE AA:", ab.fwr4_aa)

    ab.log("\n--------------")
    ab.log(" PRODUCTIVITY")
    ab.log("--------------\n")

    ab = assess_productivity(ab)
    ab.log("STOP CODON:", ab.stop_codon)
    ab.log("PRODUCTIVE:", ab.productive)
    ab.log("PRODUCTIVITY ISSUES:", ab.productivity_issues)

    return ab


ALIGNMENT_PARAMS = {
    "match": 3,
    "mismatch": -2,
    "gap_open": -35,
    "gap_extend": -1,
}



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import pytest

from ..annotation.positions import (
    get_gapped_position_from_raw,
    get_raw_position_from_aligned,
    get_raw_position_from_gapped,
)


@pytest.fixture
def gapped_sequence():
    return "ATGCATGC....ATGCATGC"


@pytest.fixture
def gapped_sequence_leading_gap():
    return "....ATGCATGCATGC"


@pytest.fixture
def gapped_sequence_trailing_gap():
    return "ATGCATGCATGC...."


@pytest.fixture
def gapped_sequence_multiple_gaps():
    return "ATGC....ATGC....ATGC"


@pytest.fixture
def aligned_sequence_no_gaps():
    return "ATGCATGCATGC"


@pytest.fixture
def aligned_sequence_with_gap():
    return "ATGC----ATGC"


@pytest.fixture
def aligned_sequence_leading_gap():
    return "----ATGCATGC"


@pytest.fixture
def aligned_sequence_trailing_gap():
    return "ATGCATGC----"


@pytest.fixture
def aligned_sequence_multiple_gaps():
    return "ATGC--GC--GC"


# ----------------------------
#      GAPPED FROM RAW
# ----------------------------


def test_get_gapped_position_from_raw(gapped_sequence):
    assert get_gapped_position_from_raw(1, gapped_sequence) == 1
    assert get_gapped_position_from_raw(2, gapped_sequence) == 2
    assert get_gapped_position_from_raw(3, gapped_sequence) == 3
    assert get_gapped_position_from_raw(4, gapped_sequence) == 4
    assert get_gapped_position_from_raw(5, gapped_sequence) == 5
    assert get_gapped_position_from_raw(6, gapped_sequence) == 6
    assert get_gapped_position_from_raw(7, gapped_sequence) == 7
    assert get_gapped_position_from_raw(8, gapped_sequence) == 8
    assert get_gapped_position_from_raw(9, gapped_sequence) == 13
    assert get_gapped_position_from_raw(10, gapped_sequence) == 14
    assert get_gapped_position_from_raw(11, gapped_sequence) == 15
    assert get_gapped_position_from_raw(12, gapped_sequence) == 16
    assert get_gapped_position_from_raw(13, gapped_sequence) == 17
    assert get_gapped_position_from_raw(14, gapped_sequence) == 18
    assert get_gapped_position_from_raw(15, gapped_sequence) == 19
    assert get_gapped_position_from_raw(16, gapped_sequence) == 20


def test_get_gapped_position_from_raw_leading_gap(gapped_sequence_leading_gap):
    assert get_gapped_position_from_raw(1, gapped_sequence_leading_gap) == 5
    assert get_gapped_position_from_raw(2, gapped_sequence_leading_gap) == 6
    assert get_gapped_position_from_raw(3, gapped_sequence_leading_gap) == 7
    assert get_gapped_position_from_raw(4, gapped_sequence_leading_gap) == 8
    assert get_gapped_position_from_raw(5, gapped_sequence_leading_gap) == 9
    assert get_gapped_position_from_raw(6, gapped_sequence_leading_gap) == 10
    assert get_gapped_position_from_raw(7, gapped_sequence_leading_gap) == 11
    assert get_gapped_position_from_raw(8, gapped_sequence_leading_gap) == 12
    assert get_gapped_position_from_raw(9, gapped_sequence_leading_gap) == 13
    assert get_gapped_position_from_raw(10, gapped_sequence_leading_gap) == 14
    assert get_gapped_position_from_raw(11, gapped_sequence_leading_gap) == 15
    assert get_gapped_position_from_raw(12, gapped_sequence_leading_gap) == 16


def test_get_gapped_position_from_raw_trailing_gap(gapped_sequence_trailing_gap):
    assert get_gapped_position_from_raw(1, gapped_sequence_trailing_gap) == 1
    assert get_gapped_position_from_raw(2, gapped_sequence_trailing_gap) == 2
    assert get_gapped_position_from_raw(3, gapped_sequence_trailing_gap) == 3
    assert get_gapped_position_from_raw(4, gapped_sequence_trailing_gap) == 4
    assert get_gapped_position_from_raw(5, gapped_sequence_trailing_gap) == 5
    assert get_gapped_position_from_raw(6, gapped_sequence_trailing_gap) == 6
    assert get_gapped_position_from_raw(7, gapped_sequence_trailing_gap) == 7
    assert get_gapped_position_from_raw(8, gapped_sequence_trailing_gap) == 8
    assert get_gapped_position_from_raw(9, gapped_sequence_trailing_gap) == 9
    assert get_gapped_position_from_raw(10, gapped_sequence_trailing_gap) == 10
    assert get_gapped_position_from_raw(11, gapped_sequence_trailing_gap) == 11
    assert get_gapped_position_from_raw(12, gapped_sequence_trailing_gap) == 12


def test_get_gapped_position_from_raw_multiple_gaps(gapped_sequence_multiple_gaps):
    assert get_gapped_position_from_raw(1, gapped_sequence_multiple_gaps) == 1
    assert get_gapped_position_from_raw(2, gapped_sequence_multiple_gaps) == 2
    assert get_gapped_position_from_raw(3, gapped_sequence_multiple_gaps) == 3
    assert get_gapped_position_from_raw(4, gapped_sequence_multiple_gaps) == 4
    assert get_gapped_position_from_raw(5, gapped_sequence_multiple_gaps) == 9
    assert get_gapped_position_from_raw(6, gapped_sequence_multiple_gaps) == 10
    assert get_gapped_position_from_raw(7, gapped_sequence_multiple_gaps) == 11
    assert get_gapped_position_from_raw(8, gapped_sequence_multiple_gaps) == 12
    assert get_gapped_position_from_raw(9, gapped_sequence_multiple_gaps) == 17
    assert get_gapped_position_from_raw(10, gapped_sequence_multiple_gaps) == 18
    assert get_gapped_position_from_raw(11, gapped_sequence_multiple_gaps) == 19
    assert get_gapped_position_from_raw(12, gapped_sequence_multiple_gaps) == 20


# ----------------------------
#      RAW FROM GAPPED
# ----------------------------


def test_get_raw_position_from_gapped(gapped_sequence):
    assert get_raw_position_from_gapped(1, gapped_sequence) == 1
    assert get_raw_position_from_gapped(2, gapped_sequence) == 2
    assert get_raw_position_from_gapped(3, gapped_sequence) == 3
    assert get_raw_position_from_gapped(4, gapped_sequence) == 4
    assert get_raw_position_from_gapped(5, gapped_sequence) == 5
    assert get_raw_position_from_gapped(6, gapped_sequence) == 6
    assert get_raw_position_from_gapped(7, gapped_sequence) == 7
    assert get_raw_position_from_gapped(8, gapped_sequence) == 8
    assert get_raw_position_from_gapped(9, gapped_sequence) == 8
    assert get_raw_position_from_gapped(10, gapped_sequence) == 8
    assert get_raw_position_from_gapped(11, gapped_sequence) == 8
    assert get_raw_position_from_gapped(12, gapped_sequence) == 8
    assert get_raw_position_from_gapped(13, gapped_sequence) == 9
    assert get_raw_position_from_gapped(14, gapped_sequence) == 10
    assert get_raw_position_from_gapped(15, gapped_sequence) == 11
    assert get_raw_position_from_gapped(16, gapped_sequence) == 12


def test_get_raw_position_from_gapped_with_sequence_start(gapped_sequence):
    # test sequence_start = 1 at multiple positions
    assert get_raw_position_from_gapped(1, gapped_sequence, sequence_start=1) == 0
    assert get_raw_position_from_gapped(2, gapped_sequence, sequence_start=1) == 1
    assert get_raw_position_from_gapped(3, gapped_sequence, sequence_start=1) == 2
    assert get_raw_position_from_gapped(4, gapped_sequence, sequence_start=1) == 3
    assert get_raw_position_from_gapped(5, gapped_sequence, sequence_start=1) == 4
    assert get_raw_position_from_gapped(6, gapped_sequence, sequence_start=1) == 5
    assert get_raw_position_from_gapped(7, gapped_sequence, sequence_start=1) == 6
    assert get_raw_position_from_gapped(8, gapped_sequence, sequence_start=1) == 7
    assert get_raw_position_from_gapped(9, gapped_sequence, sequence_start=1) == 7
    assert get_raw_position_from_gapped(10, gapped_sequence, sequence_start=1) == 7
    assert get_raw_position_from_gapped(11, gapped_sequence, sequence_start=1) == 7
    assert get_raw_position_from_gapped(12, gapped_sequence, sequence_start=1) == 7
    assert get_raw_position_from_gapped(13, gapped_sequence, sequence_start=1) == 8
    assert get_raw_position_from_gapped(14, gapped_sequence, sequence_start=1) == 9
    assert get_raw_position_from_gapped(15, gapped_sequence, sequence_start=1) == 10
    assert get_raw_position_from_gapped(16, gapped_sequence, sequence_start=1) == 11
    # test increasign sequence_start values
    assert get_raw_position_from_gapped(8, gapped_sequence, sequence_start=2) == 6
    assert get_raw_position_from_gapped(8, gapped_sequence, sequence_start=3) == 5
    assert get_raw_position_from_gapped(8, gapped_sequence, sequence_start=4) == 4
    # test increasing values at a gapped position
    assert get_raw_position_from_gapped(12, gapped_sequence, sequence_start=2) == 6
    assert get_raw_position_from_gapped(12, gapped_sequence, sequence_start=3) == 5
    assert get_raw_position_from_gapped(12, gapped_sequence, sequence_start=4) == 4


def test_get_raw_position_from_gapped_with_germline_start(gapped_sequence):
    # test germline_start = 1 at multiple positions
    assert get_raw_position_from_gapped(1, gapped_sequence, germline_start=1) == 0
    assert get_raw_position_from_gapped(2, gapped_sequence, germline_start=1) == 1
    assert get_raw_position_from_gapped(3, gapped_sequence, germline_start=1) == 2
    assert get_raw_position_from_gapped(4, gapped_sequence, germline_start=1) == 3
    assert get_raw_position_from_gapped(5, gapped_sequence, germline_start=1) == 4
    assert get_raw_position_from_gapped(6, gapped_sequence, germline_start=1) == 5
    assert get_raw_position_from_gapped(7, gapped_sequence, germline_start=1) == 6
    assert get_raw_position_from_gapped(8, gapped_sequence, germline_start=1) == 7
    assert get_raw_position_from_gapped(9, gapped_sequence, germline_start=1) == 7
    assert get_raw_position_from_gapped(10, gapped_sequence, germline_start=1) == 7
    assert get_raw_position_from_gapped(11, gapped_sequence, germline_start=1) == 7
    assert get_raw_position_from_gapped(12, gapped_sequence, germline_start=1) == 7
    assert get_raw_position_from_gapped(13, gapped_sequence, germline_start=1) == 8
    assert get_raw_position_from_gapped(14, gapped_sequence, germline_start=1) == 9
    assert get_raw_position_from_gapped(15, gapped_sequence, germline_start=1) == 10
    assert get_raw_position_from_gapped(16, gapped_sequence, germline_start=1) == 11
    # test increasign germline_start values
    assert get_raw_position_from_gapped(8, gapped_sequence, germline_start=2) == 6
    assert get_raw_position_from_gapped(8, gapped_sequence, germline_start=3) == 5
    assert get_raw_position_from_gapped(8, gapped_sequence, germline_start=4) == 4
    # test increasing values at a gapped position
    assert get_raw_position_from_gapped(12, gapped_sequence, germline_start=2) == 6
    assert get_raw_position_from_gapped(12, gapped_sequence, germline_start=3) == 5
    assert get_raw_position_from_gapped(12, gapped_sequence, germline_start=4) == 4


# ----------------------------
#      RAW FROM ALIGNED
# ----------------------------


def test_get_raw_position_from_aligned_no_gaps(aligned_sequence_no_gaps):
    assert (
        get_raw_position_from_aligned(
            position=1,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 1
    )
    assert (
        get_raw_position_from_aligned(
            position=2,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 2
    )
    assert (
        get_raw_position_from_aligned(
            position=3,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 3
    )
    assert (
        get_raw_position_from_aligned(
            position=4,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 4
    )
    assert (
        get_raw_position_from_aligned(
            position=5,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 5
    )
    assert (
        get_raw_position_from_aligned(
            position=6,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 6
    )
    assert (
        get_raw_position_from_aligned(
            position=7,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 7
    )
    assert (
        get_raw_position_from_aligned(
            position=8,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 8
    )
    assert (
        get_raw_position_from_aligned(
            position=9,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 9
    )
    assert (
        get_raw_position_from_aligned(
            position=10,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 10
    )
    assert (
        get_raw_position_from_aligned(
            position=11,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 11
    )
    assert (
        get_raw_position_from_aligned(
            position=12,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_no_gaps,
        )
        == 12
    )


def test_get_raw_position_from_aligned_with_reference_gap(
    aligned_sequence_no_gaps, aligned_sequence_with_gap
):
    assert (
        get_raw_position_from_aligned(
            position=1,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 1
    )
    assert (
        get_raw_position_from_aligned(
            position=2,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 2
    )
    assert (
        get_raw_position_from_aligned(
            position=3,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 3
    )
    assert (
        get_raw_position_from_aligned(
            position=4,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 4
    )
    assert (
        get_raw_position_from_aligned(
            position=5,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 9
    )
    assert (
        get_raw_position_from_aligned(
            position=6,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 10
    )
    assert (
        get_raw_position_from_aligned(
            position=7,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 11
    )
    assert (
        get_raw_position_from_aligned(
            position=8,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_with_gap,
        )
        == 12
    )


def test_get_raw_position_from_aligned_with_trailing_reference_gap(
    aligned_sequence_no_gaps, aligned_sequence_trailing_gap
):
    assert (
        get_raw_position_from_aligned(
            position=1,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 1
    )
    assert (
        get_raw_position_from_aligned(
            position=2,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 2
    )
    assert (
        get_raw_position_from_aligned(
            position=3,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 3
    )
    assert (
        get_raw_position_from_aligned(
            position=4,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 4
    )
    assert (
        get_raw_position_from_aligned(
            position=5,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 5
    )
    assert (
        get_raw_position_from_aligned(
            position=6,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 6
    )
    assert (
        get_raw_position_from_aligned(
            position=7,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 7
    )
    assert (
        get_raw_position_from_aligned(
            position=8,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_trailing_gap,
        )
        == 8
    )


def test_get_raw_position_from_aligned_with_leading_reference_gap(
    aligned_sequence_no_gaps, aligned_sequence_leading_gap
):
    assert (
        get_raw_position_from_aligned(
            position=1,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 5
    )
    assert (
        get_raw_position_from_aligned(
            position=2,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 6
    )
    assert (
        get_raw_position_from_aligned(
            position=3,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 7
    )
    assert (
        get_raw_position_from_aligned(
            position=4,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 8
    )
    assert (
        get_raw_position_from_aligned(
            position=5,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 9
    )
    assert (
        get_raw_position_from_aligned(
            position=6,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 10
    )
    assert (
        get_raw_position_from_aligned(
            position=7,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 11
    )
    assert (
        get_raw_position_from_aligned(
            position=8,
            aligned_sequence=aligned_sequence_no_gaps,
            aligned_reference=aligned_sequence_leading_gap,
        )
        == 12
    )



#!/usr/bin/env python
# filename: mixins.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


class LoggingMixin:
    """
    Mixin for logging messages and exceptions.

    Methods:
    --------
        log(self, *args, separator: str = " ") -> None:
            Records a log message, with each argument joined by the separator.

        exception(self, *args, separator: str = "\n") -> None:
            Records an exception, with each argument joined by the separator.

        format_log(self, separator: str = "\n") -> str:
            Formats the log, including exceptions, as a string to be written to file or stdout.

    """

    def __init__(self):
        super().__init__()
        self.logs = []
        self.exceptions = []

    def log(self, *args, separator: str = " ") -> None:
        """
        Records a log message

        Parameters:
        ----------
            *args : str
                The log message to record.

            separator : str, default: " "
                The separator to join the arguments.

        """
        log_str = separator.join([str(a) for a in args])
        self.logs.append(log_str)

    def exception(self, *args, separator: str = "\n") -> None:
        """
        Records an exception

        Parameters:
        ----------
            *args : str
                The exception to record.

            separator : str, default: "\n"
                The separator to join the arguments.

        """
        exception_str = separator.join([str(a) for a in args])
        self.exceptions.append(exception_str)

    def format_log(self, separator="\n") -> str:
        """
        Formats the log, including exceptions.

        Log formatting will only be performed on sequences that had an
        error during annotation, unless AbStar is run in debug
        mode. In debug mode, all sequences will be logged.

        Parameters:
        ----------
            separator : str, default: "\n"
                The separator to join the log messages.

        Returns:
        --------
            str: Formatted log string.

        """
        output = ""
        output += separator.join(self.logs)
        if self.exceptions:
            output += "\n\n"
            output += self._format_exceptions()
        output += "\n\n"
        return output

    def _format_exceptions(self) -> str:
        """
        Formats the exceptions as a string to be written to file or stdout.

        Returns:
        --------
            str: Formatted exceptions string.

        """
        exception_str = "\n\nEXCEPTIONS\n"
        exception_str += "----------\n\n"
        exception_str += "\n\n".join([e for e in self.exceptions])
        return exception_str


# from __future__ import absolute_import, division, print_function, unicode_literals


# class LoggingMixin(object):
#     """docstring for LoggingMixin"""

#     def __init__(self):
#         self._log = None
#         self._exceptions = None

#     @property
#     def logs(self):
#         if self._log is not None:
#             return self._log
#         return []

#     @property
#     def exceptions(self):
#         if self._exceptions is not None:
#             return self._exceptions
#         return []

#     def log(self, *args, **kwargs):
#         """
#         Records a log message
#         """
#         sep = kwargs.get("sep", " ")
#         lstring = sep.join([str(a) for a in args])
#         if self._log is None:
#             self._log = [
#                 lstring,
#             ]
#         else:
#             self._log.append(lstring)

#     def exception(self, *args, **kwargs):
#         """
#         Records an exception.
#         """
#         sep = kwargs.get("sep", "\n")
#         estring = sep.join([str(a) for a in args])
#         if self._exceptions is None:
#             self._exceptions = [
#                 estring,
#             ]
#         else:
#             self._exceptions.append(estring)

#     def format_log(self):
#         """
#         Formats the log, including exceptions.

#         Log formatting will only be performed on sequences that had an
#         error during annotation, unless AbStar is run in debug
#         mode. In debug mode, all sequences will be logged.

#         Returns:
#         --------

#             str: Formatted log string.
#         """
#         output = ""
#         output += "\n".join(self.logs)
#         if self._check_for_exceptions():
#             output += "\n\n"
#             output += self._format_exceptions()
#         output += "\n\n"
#         return output

#     def _check_for_exceptions(self):
#         if self.exceptions:
#             return True
#         if self.v is not None:
#             if self.v._exceptions:
#                 return True
#         if self.d is not None:
#             if self.d._exceptions:
#                 return True
#         if self.j is not None:
#             if self.j._exceptions:
#                 return True
#         return False

#     def _format_exceptions(self):
#         exceptions = []
#         exceptions += self.exceptions
#         estring = "\n\nEXCEPTIONS\n"
#         estring += "----------\n\n"
#         if self.v is not None:
#             exceptions += self.v.exceptions
#         if self.d is not None:
#             exceptions += self.d.exceptions
#         if self.j is not None:
#             exceptions += self.j.exceptions
#         estring += "\n\n".join([e for e in exceptions])
#         return estring



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import logging
import os
from typing import Optional

import abutils

from ..annotation.germline import get_germline_database_path


class AssignerBase:
    """
    docstring for AssignerBase
    """

    def __init__(
        self,
        output_directory: str,
        germdb_name: str,
        receptor: str,
        log_directory: str,
        logger: Optional[logging.Logger] = None,
        concise_logging: bool = False,
        debug: bool = False,
    ):
        self.output_directory = os.path.abspath(output_directory)
        self.log_directory = os.path.abspath(log_directory)
        self.receptor = receptor
        self.germdb_name = germdb_name
        self.debug = debug
        self.logger = logger if logger is not None else abutils.log.null_logger()
        self.concise_logging = concise_logging
        self.germdb_path = get_germline_database_path(germdb_name, receptor)
        self.to_delete = []  # files that should be deleted during cleanup

        abutils.io.make_dir(self.output_directory)

    def __call__(self, sequence_file: str) -> str:
        """
        All classes subclassing ``AssignerBase`` must implement this method.

        The only expected argument is `sequence_file`, which is a single FASTA or FASTQ-formatted
        file containing input sequences. Gzipped files are supported.

        The return value should be the path to a parquet-formatted file containing the following
        columns:
          * ``sequence_id``: a unique identifier for each input sequence.
          * ``sequence``: the input sequence.
          * ``quality``: the quality score for each input sequence.
          * ``rev_comp``: a boolean indicating whether the input sequence is in the reverse_complement orientation
          * ``v_call``: the V-gene call for each input sequence.
          * ``v_support``: the V-gene evalue for each input sequence.
          * ``d_call``: the D-gene call for each input sequence.
          * ``d_support``: the D-gene evalue for each input sequence.
          * ``j_call``: the J-gene call for each input sequence.
          * ``j_support``: the J-gene evalue for each input sequence.

        """
        raise NotImplementedError("Subclasses must implement this method")

    def cleanup(self):
        abutils.io.delete_files(self.to_delete)
        self.to_delete = []

    # @staticmethod
    # def get_germdb_path(germdb_name: str, receptor: str = "bcr") -> str:
    #     """
    #     Get the path to a germline database.

    #     Parameters
    #     ----------
    #     germdb_name : str
    #         The name of the germline database.

    #     receptor : str, default: "bcr"
    #         The receptor type. Options are "bcr" and "tcr".

    #     Returns
    #     -------
    #     str
    #         The path to the germline database.
    #     """
    #     germdb_name = germdb_name.lower()
    #     receptor = receptor.lower()
    #     if receptor not in ["bcr", "tcr"]:
    #         raise ValueError(f"Receptor type {receptor} not supported")

    #     # check the addon directory first
    #     addon_dir = os.path.expanduser(f"~/.abstar/germline_dbs/{receptor}")
    #     if os.path.isdir(addon_dir):
    #         if germdb_name.lower() in [
    #             os.path.basename(d[0]) for d in os.walk(addon_dir)
    #         ]:
    #             return os.path.join(addon_dir, f"{receptor}/{germdb_name}")

    #     # if a user-generated DB isn't found, use the built-in DB
    #     curr_dir = os.path.dirname(os.path.abspath(__file__))
    #     germdb_path = os.path.join(curr_dir, f"germline_dbs/{receptor}/{germdb_name}")
    #     if not os.path.exists(germdb_path):
    #         raise FileNotFoundError(
    #             f"Germline database {germdb_name} for receptor {receptor} not found"
    #         )
    #     return germdb_path


##
##
##
##
##
##


# class BaseAssigner:
#     """
#     ``BaseAssigner`` provides an abstract base class for custom VDJ Assigners.

#     Overview
#     --------

#     Several different tools exist for inferring the germline components of recombined
#     antibody sequences. Due to their differing design goals, these tools often have
#     different priorities -- some are optimized for speed with a small (often
#     negligible) accuracy penalty, while others define recombination junctions with high
#     precision at the cost of speed. Because there isn't a single germline assignment tool
#     that is optimal for every use case, the germline assignment component of AbStar has been
#     designed to be modular and replaceable, allowing the relatively straightforward addition
#     of new germline assignment tools and the selection of the desired Assigner at runtime.
#     AbStar's default Assigner (``blastn``) identifies V- and J-genes using BLASTn and identifies
#     D-genes using rapid Smith-Waterman local alignment.

#     The basic purpose of any custom Assigner class is to accept sequences and
#     produce ``VDJ`` objects. ``VDJ`` objects package the input sequence together with
#     ``GermlineSegment`` objects that contain information about V, D and/or J assignments.
#     Following germline assignment by the Assigner, addtional annotation will be performed
#     by AbStar. The ``VDJ`` objects provide a common interface by which various Assigners
#     can communicate with the additional sequence annotation components in AbStar. This additional
#     annotation is consistent regardless of the Assigner (as is the output schema), and provides
#     a unified method by which different germline Assigners can be used across different projects
#     while maintaining schematic compatibility with downstream analysis tools.

#     ``BaseAssigner`` provides the following attributes:

#       - ``assigned``: a list designed to contain ``VDJ`` objects for which
#                     successful V(D)J assignment has been completed. These
#                     ``VDJ`` objects will be additionally annotated by AbStar
#                     following assignment. Exception/log information in these
#                     ``VDJ`` objects will only be written to the log file if AbStar
#                     is run in debug mode.

#       - ``unassigned``: a list designed to contain ``VDJ`` objects for which
#                       V(D)J assigjment was unsuccessful. These ``VDJ`` objects
#                       will not be additionally annotated by AbStar and any exception/log
#                       information contained in these ``VDJ`` objects will be logged
#                       (even if AbStar was not run in debug mode).

#       - ``germline_directory``: path to the directory containing germline databases.
#                               Germline DBs are heirarchically grouped into separate folders
#                               first by species, then by the Assigner for which they were
#                               designed (``blastn``, for example). Within each folder, germline
#                               DB files are named as ``{segment}.fasta``, so the Blast V-gene
#                               DB for human would be ``human/blast/v.fasta``. Note that this will
#                               change as AbStar is updated to annotate TCRs in addition to antibodies

#       - ``binary_directory``: path to the directory containing Assigner binaries. Binaries
#                             are named as ``{binary}_{system}``, where ``system`` is the
#                             lowercase output from platform.system(). For example, the
#                             Blastn binary for OSX would be at ``binary_directory/blastn_darwin``.
#                             Some assigners require more than a simple binary for execution (Partis,
#                             for example). In this case, all required elements will be placed into
#                             a folder within the binary directory, named using the same convention
#                             as binaries. So the folder containing Linux-compiled Partis components
#                             will be ``binary_directory/partis_linux``. The main Partis executable
#                             would then be located at ``binary_directory/partis_linux/bin/partis``.

#     In order to build a custom assignment class, you simply need to subclass BaseAssigner, implement
#     your assigner, and register your assigner in abstar.assigners.registry.ASSIGNERS. Obviously, this
#     is in addition to adding any required binaries and specialized germline database formats. AbStar
#     prefers to incorporate assigner binaries if at all possible (as opposed to relying on system
#     installations) to smooth the process of installing AbStar and eliminate the need for users to
#     install several additional programs before having a fully operational AbStar installation.

#     There are two critical elements to the design of your assigner class. First, ensure that
#     you call the BaseAssigner's ``__init__()`` at the start of your class's ``__init__()``. Second,
#     you must implement the  ``__call__()`` method.


#     ``__init__()``
#     --------------

#     Your custom assigner class's ``__init__()`` must accept a single argument (``species``, in
#     addition to the required ``self``). You must then call the BaseAssigner's ``__init__()`` and pass
#     the ``species`` argument. The simplest was to do this is by using ``super()``, like this:
#     ``super(MyAssigner, self).__init__(species)``, where ``MyAssigner`` is your assigner class.


#     Implementing the ``__call__()`` method
#     --------------------------------------

#       - ``__call__()`` must accept two arguments. The first (``sequence_file``) is a FASTA- or
#         FASTQ-formatted file of input sequences, depending on the user-provided input. The second argument
#         (``file_format``) defines the format of ``sequence_file``, either ``'fasta'`` or ``'fastq'``. If
#         a FASTQ file is required by your assigner, ensure that ``__call__()`` raises the appropriate exception
#         (and provides a sufficiently clear description of the problem) if a FASTA-formatted file is provided.

#       - The result of ``__call__()`` should be the generation of a ``VDJ`` object for each input sequence
#         (more on this below). The ``VDJ`` objects should be appended to either the ``assigned`` or ``unassigned``
#         attribute, depending on the outcome of the V(D)J assignment. A high-level overview of how ``__call__()``
#         should work is as follows::

#             def __call__(self, seqs, species):
#                 for seq in seqs:
#                     vdj = self.assign_vdj(seq, species)
#                     if self.is_properly_assigned(vdj):
#                         self.assigned.append(vdj)
#                     else:
#                         self.unassigned.append(vdj)

#         In this case, the output of ``assign_vdj()`` should be a ``VDJ`` object.


#     V(D)J assignment
#     ----------------

#     AbStar defines ``VDJ`` objects, which are designed to package the input sequence and ``GermlineSegment``
#     objects that provide information about germline assignment for V, D and/or J genes. ``VDJ`` objects must
#     contain the input sequence and may also contain information about V, D and J assignments. While the
#     sequence is required, the germline assignments are optional, to allow creation of ``VDJ`` objects for
#     light chains (which lack a D-gene) and/or for unsuccessful assignments. Additionally, this allows
#     flexibility in designing the ``__call__()`` method. Assigners can either instantiate a ``VDJ`` object at
#     the start of ``__call__()`` and attach ``GermlineSegment``s as they are computed or they can instantiate
#     ``VDJ`` objects upon completion of V(D)J assignment using the input sequences and computed
#     ``GermlineSegment``s as instantiation arguments. Because ``VDJ`` objects contain built-in logging
#     capability, such objects should be created for each input sequence even in the case of unsuccessful
#     assignments so that assignment issues can be logged and reported.

#     ``VDJ`` objects include a logging method: ``log()``. Using ``log()`` is similar to using Python's
#     builtin ``print()`` function. Multiple arguments to ``log()`` will be space-delimited, and a ``sep``
#     keyword allows you to provide an alternate delimiter. Log messages are typically used to record useful
#     information or to note the occurance of errors/exceptions that are caught and handled by the Assigner.
#     It is also a useful means by which to note the reason a specific input sequence is unable to be fully
#     processed (because it isn't a valid antibody sequence, etc). By convention, AbStar log entries should
#     contain the name of the log event (in all caps) followed by a colon and additional log information.
#     An example ``log()`` call in the case of a V(D)J assignment that was aborted because the V-gene alignment
#     score didn't meet the alignment quality threshold::

#         for seq in seqs:
#             vdj = VDJ(seq)
#             v = assign_v_gene(seq)
#             if v.score < threshold:
#                 vdj.log('V-GENE ASSIGNMENT ERROR:', 'Alignment score ({}) was too low'.format(v.score))
#                 continue

#     ``VDJ`` objects also contain a mechanism for logging exceptions: ``exception()``. This is typically used
#     to log unexpected exceptions (ie, those that are not caught and handled by the Assigner). It is useful
#     to provide some information about when/where in the Assigner the exception occured, as well as a formatted
#     version of the exception traceback. An example of ``exceptions()`` being used to log an exception that
#     occurs during V-gene assignment::

#         vdj = VDJ(seq)
#         try:
#             v = assign_v_gene(seq)
#         except:
#             vdj.exception('V-GENE ASSIGNMENT EXCEPTION:', traceback.format_exc())

#     One of the primary benefits of recording log and exception information through the ``VDJ`` objects is
#     that it allows for much easier formatting of such events when multiple sequences are being processed
#     in parallel. Using Python's ``logging`` utility would result in log events being logged as they occur,
#     which means that multiple log entries for the same input sequence may be far apart in the resulting log
#     file. Packaging log and exception information with each ``VDJ`` object alows AbStar to create a much
#     more useful log file in which all information related to a single input sequence is found in a single
#     log location.


#     ``GermlineSegment`` objects
#     ---------------------------

#     ``GermlineSegment`` objects contain information about the assigned germline gene segment (V, D or J).
#     Instantiation of a ``GermlineSegment`` object requires only the name of the germline gene segment (in
#     IMGT format, like 'IGHV3-23*01') and the database name. Several other optional arguments can be included
#     at instantiation. ``score`` is the assignment score, typically an ``int`` or ``float``. ``strand``
#     indicates the strand orientation of the input sequence (``'+'`` or ``'-'``). ``others`` is a list
#     of additional high scoring ``GermlineSegment`` objects. ``assigner_name`` is the name of the custom
#     Assigner (as a ``str``), and will be converted to lowercase before recording in AbStar output.

#     ``GermlineSegment`` objects also provide a few additional public attributes that may be desirable
#     for certain custom Assigners, especially those that have been specifically designed to accurately
#     define recombination regions. ``query_start``, ``query_end``, ``germline_start`` and ``germline_end``
#     allow the Assigner to define the start and end points of the germline alignment for both the query
#     and germline sequences. Note that both ``query_start`` and ``query_start`` and ``query_end`` MUST
#     be relative to the complete query sequence in the ``'+'`` orientation (in other words, V-gene at
#     the 5' end and J-gene at the 3' end). AbStar realigns all assigned germline genes using alignment
#     parameters designed to accurately identify somatic mutations as well as somatic mutation-induced
#     indels. If the additional assignment position attributes are provided by the Assigner, AbStar will
#     force their use during realignment. If not, AbStar will use the assigned germline genes for realignment
#     and will define the start and end points of the germline alignment during realignment.


#     Example
#     -------

#     The following is an example of a custom Assigner class, named MyAssigner. It uses BioPython's SeqIO
#     to parse the input file, and because BioPython is a dependency of AbStar, you can assume that BioPython
#     will be present on any system running AbStar. All custom Assigners should be located in the ``assigners``
#     directory::

#         import os
#         import platform
#         import traceback

#         from abstar.assigners.assigner import BaseAssigner
#         from abstar.core.vdj import VDJ
#         from abstar.core.germline import GermlineSegment

#         from Bio import SeqIO

#         class MyAssigner(BaseAssigner):

#             def __init__(self, db_name):
#                 super(MyAssigner, self).__init__(db_name)
#                 self.binary = os.path.join(self.binary_directory, 'mybinary_{}'.format(platform.system()))

#             def __call__(self, sequence_file, file_format):
#                 for sequence in SeqIO.parse(open(sequence_file, 'r'), file_format):
#                     seq = Sequence(sequence)
#                     vdj = VDJ(seq)

#                     # V-gene assignment
#                     # The VDJ object (rather than just the sequence) is passed to the assignment
#                     # function for three reasons:
#                     #    1) it contains the sequence, so we don't need to pass it separately
#                     #    2) it allows logging during germline assignment, via vdj.log()
#                     #    3) subsequent assignments will have access to information about
#                     #       previous assignments (J-gene assignment operations have access
#                     #       to information about the V-gene assignment, etc.)
#                     try:
#                         vdj.v = self.assign_germline(vdj, 'V')
#                     except:
#                         vdj.exception('V-ASSIGNMENT:', traceback.format_exc())

#                     # J-gene assignment
#                     try:
#                         vdj.j = self.assign_germline(vdj, 'J')
#                     except:
#                         vdj.exception('J-ASSIGNMENT:', traceback.format_exc())

#                     # D-gene assignment
#                     try:
#                         vdj.d = self.assign_germline(vdj, 'D')
#                     except:
#                         vdj.exception('D-ASSIGNMENT:', traceback.format_exc())

#                     return vdj

#             def assign_germline(self, vdj, segment):
#                 germ_db = os.path.join(self.germline_directory, '{}/ungapped/{}.fasta'.format(self.db_name,
#                                                                                               segment.lower()))

#                 # do stuff to assign the germline gene, using the species-appropriate germline DB
#                 # ...
#                 # ...
#                 # ...
#                 # germs is a list of germline genes, ordered by score (highest first)

#                 if germs[0].score < threshold:
#                     vdj.log('{}-ASSIGNMENT ERROR:'.format(segment),
#                             'Score ({}) is too low'.format(germs[0].score))
#                     return None
#                 others = [GermlineSegment(germ.name, self.db_name, score=germ.score) for germ in germs[1:6]]
#                 return GermlineSegment(germs[0].name, self.db_name, score=germs[0].score, others=others)

#     """

#     def __init__(self, db_name, receptor):
#         super(BaseAssigner, self).__init__()
#         self.name = self.__class__.__name__.lower()
#         self.species = db_name
#         self.db_name = db_name
#         self.receptor = receptor.lower()
#         self._assigned = None
#         self._unassigned = None
#         self._germline_directory = None
#         self._binary_directory = None

#     def __call__(self, sequence_file: str, file_format: str):
#         # __call__ should be implemented by subclasses
#         pass

#     @property
#     def germline_directory(self):
#         """
#         returns a path to the directory containing the germline database for the species and receptor
#         """
#         if self._germline_directory is None:
#             self._germline_directory = get_germline_database_directory(
#                 self.db_name, self.receptor
#             )
#         return self._germline_directory

#     @germline_directory.setter
#     def germline_directory(self, directory):
#         self._germline_directory = directory

#     @property
#     def binary_directory(self):
#         """
#         returns a path to the directory containing the binary for the assigner
#         """
#         if self._binary_directory is None:
#             mod_dir = os.path.dirname(os.path.abspath(__file__))
#             self._binary_directory = os.path.join(mod_dir, "bin")
#         return self._binary_directory

#     @property
#     def assigned(self):
#         """
#         returns a list of VDJ objects that were successfully assigned
#         """
#         if self._assigned is None:
#             self._assigned = []
#         return self._assigned

#     @assigned.setter
#     def assigned(self, assigned):
#         self._assigned = assigned

#     @property
#     def unassigned(self):
#         """
#         returns a list of VDJ objects that were not successfully assigned
#         """
#         if self._unassigned is None:
#             self._unassigned = []
#         return self._unassigned

#     @unassigned.setter
#     def unassigned(self, unassigned):
#         self._unassigned = unassigned



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

from typing import Optional

from .antibody import Antibody
from .positions import get_gapped_position_from_raw


def annotate_mutations(
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
    ab: Optional[Antibody] = None,
    debug: bool = False,
) -> Antibody:
    """
    Annotates mutations in an antibody germline gene region. The notation abstar uses for mutations is:

        ``"123:A>T"`` (if the mutation is a single nucleotide change)

    .. note::
        The mutation position in abstar notation is the IMGT position of the mutation,
        which is 1-indexed

    Parameters
    ----------
    aligned_sequence : str
        The aligned sequence of the query antibody region.

    aligned_germline : str
        The aligned sequence of the reference germline gene.

    gapped_germline : str
        The IMGT-gapped sequence of the reference germline gene.

    germline_start : int
        The starting position of the aligned germline sequence relative to the start of the
        complete germline gene.

    ab : Antibody, optional
        If provided, the ``Antibody`` object is only used for logging purposes.

    debug : bool, optional
        If ``True``, the mutation annotation process is logged to the ``Antibody`` object's
        logger.

    Returns
    -------
    mutations : list[str]
        List of mutations in abstar notation.

    """
    mutations = []
    raw_position = germline_start
    if ab is not None:
        ab.log("ALIGNED SEQUENCE:", aligned_sequence)
        ab.log("ALIGNED GERMLINE:", aligned_germline)
        ab.log("GAPPED GERMLINE:", gapped_germline)
        ab.log("GERMLINE START:", germline_start)
        ab.log("GERM      SEQ       RAW       IMGT      MUTATION  ")
        # ab.log("--------------------------------------------------")
    for s, g in zip(aligned_sequence, aligned_germline):
        log_str = f"{g}{' ' * (10 - len(g))}{s}{' ' * (10 - len(s))}"
        if s == g:
            # sequence and germline match -- no mutation and increment the position
            raw_position += 1
            log_str += f"{raw_position}{' ' * (10 - len(str(raw_position)))}"
            if debug:
                imgt_position = get_gapped_position_from_raw(
                    raw_position, gapped_germline
                )
                log_str += f"{imgt_position}{' ' * (10 - len(str(imgt_position)))}"
        elif g == "-":
            # sequence has insertion -- no mutation and don't increment the position
            log_str += f"{raw_position}{' ' * (10 - len(str(raw_position)))}"
            if debug:
                imgt_position = get_gapped_position_from_raw(
                    raw_position, gapped_germline
                )
                log_str += f"{imgt_position}{' ' * (10 - len(str(imgt_position)))}"
        elif s == "-":
            # sequence has a deletion -- no mutation but increment the position
            raw_position += 1
        else:
            # yay, we found a mutation!
            raw_position += 1
            imgt_position = get_gapped_position_from_raw(raw_position, gapped_germline)
            mutation = f"{imgt_position}:{g}>{s}"
            mutations.append(mutation)
            log_str += f"{raw_position}{' ' * (10 - len(str(raw_position)))}"
            log_str += f"{imgt_position}{' ' * (10 - len(str(imgt_position)))}"
            log_str += f"{mutation}"
        if ab is not None:
            ab.log(log_str)
    return mutations


def annotate_v_mutations(
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
    is_aa: bool,
    ab: Antibody,
    debug: bool = False,
) -> Antibody:
    """
    Annotates mutations in an antibody germline gene region. The notation abstar uses for mutations is:

        ``"123:A>T"`` (if the mutation is a single nucleotide change)

    .. note::
        The mutation position in abstar notation is the IMGT position of the mutation,
        which is 1-indexed

    Parameters
    ----------
    aligned_sequence : str
        The aligned sequence of the query antibody region.

    aligned_germline : str
        The aligned sequence of the reference germline gene.

    gapped_germline : str
        The IMGT-gapped sequence of the reference germline gene.

    germline_start : int
        The starting position of the aligned germline sequence relative to the start of the
        complete germline gene.

    is_aa : bool
        If ``True``, all input sequences contain amino acids (not nucleotides).

    ab : Antibody
        Antibody object to update with annotation information. The following ``Antibody``
        properties are updated:

        - ``v_mutations`` (if ``is_aa=False``)
        - ``v_mutation_count`` (if ``is_aa=False``)

        - ``v_mutations_aa`` (if ``is_aa=True``)
        - ``v_mutation_count_aa`` (if ``is_aa=True``)

    debug : bool, optional
        If ``True``, the mutation annotation process is logged to the ``Antibody`` object's
        logger.

    Returns
    -------
    ab : Antibody
        Antibody object with updated annotation information.

    """
    mutations = annotate_mutations(
        aligned_sequence=aligned_sequence,
        aligned_germline=aligned_germline,
        gapped_germline=gapped_germline,
        germline_start=germline_start,
        ab=ab,
        debug=debug,
    )
    if is_aa:
        ab.v_mutations_aa = "|".join(mutations)
        ab.v_mutation_count_aa = len(mutations)
    else:
        ab.v_mutations = "|".join(mutations)
        ab.v_mutation_count = len(mutations)
    return ab


def annotate_c_mutations(
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
    is_aa: bool,
    ab: Antibody,
    debug: bool = False,
) -> Antibody:
    """
    Annotates mutations in an antibody germline gene region. The notation abstar uses for mutations is:

        ``"123:A>T"`` (if the mutation is a single nucleotide change)

    .. note::
        The mutation position in abstar notation is the IMGT position of the mutation,
        which is 1-indexed

    Parameters
    ----------
    aligned_sequence : str
        The aligned sequence of the query antibody region.

    aligned_germline : str
        The aligned sequence of the reference germline gene.

    gapped_germline : str
        The IMGT-gapped sequence of the reference germline gene.

    germline_start : int
        The starting position of the aligned germline sequence relative to the start of the
        complete germline gene.

    is_aa : bool
        If ``True``, all input sequences contain amino acids (not nucleotides).

    ab : Antibody
        Antibody object to update with annotation information. The following ``Antibody``
        properties are updated:

        - ``c_mutations`` (if ``is_aa=False``)
        - ``c_mutation_count`` (if ``is_aa=False``)

        - ``c_mutations_aa`` (if ``is_aa=True``)
        - ``c_mutation_count_aa`` (if ``is_aa=True``)

    debug : bool, optional
        If ``True``, the mutation annotation process is logged to the ``Antibody`` object's
        logger.

    Returns
    -------
    ab : Antibody
        Antibody object with updated annotation information.

    """
    mutations = annotate_mutations(
        aligned_sequence=aligned_sequence,
        aligned_germline=aligned_germline,
        gapped_germline=gapped_germline,
        germline_start=germline_start,
        ab=ab,
        debug=debug,
    )
    if is_aa:
        ab.c_mutations_aa = "|".join(mutations)
        ab.c_mutation_count_aa = len(mutations)
    else:
        ab.c_mutations = "|".join(mutations)
        ab.c_mutation_count = len(mutations)
    return ab
